# Copyright 2014 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral.db.v2 import api as db_api
from mistral.engine import action_queue
from mistral.engine import base
from mistral import expressions
from mistral.services import scheduler
from mistral.utils import wf_trace
from mistral.workflow import data_flow
from mistral.workflow import states

import six

_CONTINUE_TASK_PATH = 'mistral.engine.policies._continue_task'
_COMPLETE_TASK_PATH = 'mistral.engine.policies._complete_task'


def _log_task_delay(task_ex, delay_sec):
    wf_trace.info(
        task_ex,
        "Task '%s' [%s -> %s, delay = %s sec]" %
        (task_ex.name, task_ex.state, states.RUNNING_DELAYED, delay_sec)
    )


def build_policies(policies_spec, wf_spec):
    task_defaults = wf_spec.get_task_defaults()
    wf_policies = task_defaults.get_policies() if task_defaults else None

    if not (policies_spec or wf_policies):
        return []

    return construct_policies_list(policies_spec, wf_policies)


def get_policy_factories():
    return [
        build_pause_before_policy,
        build_wait_before_policy,
        build_wait_after_policy,
        build_retry_policy,
        build_timeout_policy,
        build_concurrency_policy
    ]


def construct_policies_list(policies_spec, wf_policies):
    policies = []

    for factory in get_policy_factories():
        policy = factory(policies_spec)

        if wf_policies and not policy:
            policy = factory(wf_policies)

        if policy:
            policies.append(policy)

    return policies


def build_wait_before_policy(policies_spec):
    if not policies_spec:
        return None

    wait_before = policies_spec.get_wait_before()

    if (isinstance(wait_before, six.string_types) or wait_before > 0):
        return WaitBeforePolicy(wait_before)
    else:
        return None


def build_wait_after_policy(policies_spec):
    if not policies_spec:
        return None

    wait_after = policies_spec.get_wait_after()

    if (isinstance(wait_after, six.string_types) or wait_after > 0):
        return WaitAfterPolicy(wait_after)
    else:
        return None


def build_retry_policy(policies_spec):
    if not policies_spec:
        return None

    retry = policies_spec.get_retry()

    if not retry:
        return None

    return RetryPolicy(
        retry.get_count(),
        retry.get_delay(),
        retry.get_break_on(),
        retry.get_continue_on()
    )


def build_timeout_policy(policies_spec):
    if not policies_spec:
        return None

    timeout_policy = policies_spec.get_timeout()

    if (isinstance(timeout_policy, six.string_types) or timeout_policy > 0):
        return TimeoutPolicy(timeout_policy)
    else:
        return None


def build_pause_before_policy(policies_spec):
    if not policies_spec:
        return None

    pause_before_policy = policies_spec.get_pause_before()

    return (PauseBeforePolicy(pause_before_policy)
            if pause_before_policy else None)


def build_concurrency_policy(policies_spec):
    if not policies_spec:
        return None

    concurrency_policy = policies_spec.get_concurrency()

    return (ConcurrencyPolicy(concurrency_policy)
            if concurrency_policy else None)


def _ensure_context_has_key(runtime_context, key):
    if not runtime_context:
        runtime_context = {}

    if key not in runtime_context:
        runtime_context.update({key: {}})

    return runtime_context


class WaitBeforePolicy(base.TaskPolicy):
    _schema = {
        "properties": {
            "delay": {"type": "integer"}
        }
    }

    def __init__(self, delay):
        self.delay = delay

    def before_task_start(self, task_ex, task_spec):
        super(WaitBeforePolicy, self).before_task_start(task_ex, task_spec)

        context_key = 'wait_before_policy'

        runtime_context = _ensure_context_has_key(
            task_ex.runtime_context,
            context_key
        )

        task_ex.runtime_context = runtime_context

        policy_context = runtime_context[context_key]

        if policy_context.get('skip'):
            # Unset state 'RUNNING_DELAYED'.
            wf_trace.info(
                task_ex,
                "Task '%s' [%s -> %s]"
                % (task_ex.name, states.RUNNING_DELAYED, states.RUNNING)
            )

            task_ex.state = states.RUNNING

            return

        if task_ex.state != states.IDLE:
            policy_context.update({'skip': True})

            _log_task_delay(task_ex, self.delay)

            task_ex.state = states.RUNNING_DELAYED

            scheduler.schedule_call(
                None,
                _CONTINUE_TASK_PATH,
                self.delay,
                task_ex_id=task_ex.id,
            )


class WaitAfterPolicy(base.TaskPolicy):
    _schema = {
        "properties": {
            "delay": {"type": "integer"}
        }
    }

    def __init__(self, delay):
        self.delay = delay

    def after_task_complete(self, task_ex, task_spec):
        super(WaitAfterPolicy, self).after_task_complete(task_ex, task_spec)

        context_key = 'wait_after_policy'

        runtime_context = _ensure_context_has_key(
            task_ex.runtime_context,
            context_key
        )

        task_ex.runtime_context = runtime_context

        policy_context = runtime_context[context_key]

        if policy_context.get('skip'):
            # Skip, already processed.
            return

        policy_context.update({'skip': True})

        _log_task_delay(task_ex, self.delay)

        end_state = task_ex.state
        end_state_info = task_ex.state_info

        # TODO(rakhmerov): Policies probably needs to have tasks.Task
        # interface in order to change manage task state safely.
        # Set task state to 'DELAYED'.
        task_ex.state = states.RUNNING_DELAYED
        task_ex.state_info = (
            'Suspended by wait-after policy for %s seconds' % self.delay
        )

        # Schedule to change task state to RUNNING again.
        scheduler.schedule_call(
            None,
            _COMPLETE_TASK_PATH,
            self.delay,
            task_ex_id=task_ex.id,
            state=end_state,
            state_info=end_state_info
        )


class RetryPolicy(base.TaskPolicy):
    _schema = {
        "properties": {
            "delay": {"type": "integer"},
            "count": {"type": "integer"}
        }
    }

    def __init__(self, count, delay, break_on, continue_on):
        self.count = count
        self.delay = delay
        self.break_on = break_on
        self._continue_on_clause = continue_on

    def after_task_complete(self, task_ex, task_spec):
        """Possible Cases:

        1. state = SUCCESS
           if continue_on is not specified,
           no need to move to next iteration;
           if current:count achieve retry:count then policy
           breaks the loop (regardless on continue-on condition);
           otherwise - check continue_on condition and if
           it is True - schedule the next iteration,
           otherwise policy breaks the loop.
        2. retry:count = 5, current:count = 2, state = ERROR,
           state = IDLE/DELAYED, current:count = 3
        3. retry:count = 5, current:count = 4, state = ERROR
        Iterations complete therefore state = #{state}, current:count = 4.
        """
        super(RetryPolicy, self).after_task_complete(task_ex, task_spec)

        # TODO(m4dcoder): If the task_ex.action_executions and
        # task_ex.workflow_executions collection are not called,
        # then the retry_no in the runtime_context of the task_ex will not
        # be updated accurately. To be exact, the retry_no will be one
        # iteration behind.
        ex = task_ex.executions  # noqa

        context_key = 'retry_task_policy'

        runtime_context = _ensure_context_has_key(
            task_ex.runtime_context,
            context_key
        )

        wf_ex = task_ex.workflow_execution

        ctx_view = data_flow.ContextView(
            data_flow.evaluate_task_outbound_context(task_ex),
            wf_ex.context,
            wf_ex.input
        )

        continue_on_evaluation = expressions.evaluate(
            self._continue_on_clause,
            ctx_view
        )

        task_ex.runtime_context = runtime_context

        state = task_ex.state

        if not states.is_completed(state) or states.is_cancelled(state):
            return

        policy_context = runtime_context[context_key]

        retry_no = 0

        if 'retry_no' in policy_context:
            retry_no = policy_context['retry_no']
            del policy_context['retry_no']

        retries_remain = retry_no < self.count

        stop_continue_flag = (
            task_ex.state == states.SUCCESS and
            not self._continue_on_clause
        )

        stop_continue_flag = (
            stop_continue_flag or
            (self._continue_on_clause and not continue_on_evaluation)
        )

        break_triggered = (
            task_ex.state == states.ERROR and
            self.break_on
        )

        if not retries_remain or break_triggered or stop_continue_flag:
            return

        _log_task_delay(task_ex, self.delay)

        data_flow.invalidate_task_execution_result(task_ex)
        task_ex.state = states.RUNNING_DELAYED

        policy_context['retry_no'] = retry_no + 1
        runtime_context[context_key] = policy_context

        scheduler.schedule_call(
            None,
            _CONTINUE_TASK_PATH,
            self.delay,
            task_ex_id=task_ex.id,
        )


class TimeoutPolicy(base.TaskPolicy):
    _schema = {
        "properties": {
            "delay": {"type": "integer"}
        }
    }

    def __init__(self, timeout_sec):
        self.delay = timeout_sec

    def before_task_start(self, task_ex, task_spec):
        super(TimeoutPolicy, self).before_task_start(task_ex, task_spec)

        scheduler.schedule_call(
            None,
            'mistral.engine.policies._fail_task_if_incomplete',
            self.delay,
            task_ex_id=task_ex.id,
            timeout=self.delay
        )

        wf_trace.info(
            task_ex,
            "Timeout check scheduled [task=%s, timeout(s)=%s]." %
            (task_ex.id, self.delay)
        )


class PauseBeforePolicy(base.TaskPolicy):
    _schema = {
        "properties": {
            "expr": {"type": "boolean"}
        }
    }

    def __init__(self, expression):
        self.expr = expression

    def before_task_start(self, task_ex, task_spec):
        super(PauseBeforePolicy, self).before_task_start(task_ex, task_spec)

        if not self.expr:
            return

        wf_trace.info(
            task_ex,
            "Workflow paused before task '%s' [%s -> %s]" %
            (task_ex.name, task_ex.workflow_execution.state, states.PAUSED)
        )

        task_ex.workflow_execution.state = states.PAUSED
        task_ex.state = states.IDLE


class ConcurrencyPolicy(base.TaskPolicy):
    _schema = {
        "properties": {
            "concurrency": {"type": "integer"},
        }
    }

    def __init__(self, concurrency):
        self.concurrency = concurrency

    def before_task_start(self, task_ex, task_spec):
        super(ConcurrencyPolicy, self).before_task_start(task_ex, task_spec)

        # This policy doesn't do anything except validating "concurrency"
        # property value and setting a variable into task runtime context.
        # This variable is then used to define how many action executions
        # may be started in parallel.
        context_key = 'concurrency'

        runtime_context = _ensure_context_has_key(
            task_ex.runtime_context,
            context_key
        )

        runtime_context[context_key] = self.concurrency
        task_ex.runtime_context = runtime_context


@action_queue.process
def _continue_task(task_ex_id):
    from mistral.engine import task_handler

    with db_api.transaction():
        task_handler.continue_task(db_api.get_task_execution(task_ex_id))


@action_queue.process
def _complete_task(task_ex_id, state, state_info):
    from mistral.engine import task_handler

    with db_api.transaction():
        task_handler.complete_task(
            db_api.get_task_execution(task_ex_id),
            state,
            state_info
        )


@action_queue.process
def _fail_task_if_incomplete(task_ex_id, timeout):
    from mistral.engine import task_handler

    with db_api.transaction():
        task_ex = db_api.get_task_execution(task_ex_id)

        if not states.is_completed(task_ex.state):
            msg = 'Task timed out [timeout(s)=%s].' % timeout

            task_handler.complete_task(
                db_api.get_task_execution(task_ex_id),
                states.ERROR,
                msg
            )



# Copyright 2016 - Brocade Communications Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_config import cfg
from oslo_log import log as logging
import pprint
import requests

from mistral import auth


LOG = logging.getLogger(__name__)

CONF = cfg.CONF


class KeycloakAuthHandler(auth.AuthHandler):

    def authenticate(self, req):
        realm_name = req.headers.get('X-Project-Id')

        # NOTE(rakhmerov): There's a special endpoint for introspecting
        # access tokens described in OpenID Connect specification but it's
        # available in KeyCloak starting only with version 1.8.Final so we have
        # to use user info endpoint which also takes exactly one parameter
        # (access token) and replies with error if token is invalid.
        user_info_endpoint = (
            "%s/realms/%s/protocol/openid-connect/userinfo" %
            (CONF.keycloak_oidc.auth_url, realm_name)
        )

        access_token = req.headers.get('X-Auth-Token')

        resp = requests.get(
            user_info_endpoint,
            headers={"Authorization": "Bearer %s" % access_token},
            verify=not CONF.keycloak_oidc.insecure
        )

        resp.raise_for_status()

        LOG.debug(
            "HTTP response from OIDC provider: %s" %
            pprint.pformat(resp.json())
        )



# Copyright 2015 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Add database indices

Revision ID: 009
Revises: 008
Create Date: 2015-11-25 19:06:14.975474

"""

# revision identifiers, used by Alembic.
revision = '009'
down_revision = '008'

from alembic import op
from sqlalchemy.engine import reflection


def upgrade():

    inspector = reflection.Inspector.from_engine(op.get_bind())

    op.create_index(
        'action_definitions_v2_action_class',
        'action_definitions_v2',
        ['action_class'],
        unique=False
    )

    op.create_index(
        'action_definitions_v2_is_system',
        'action_definitions_v2',
        ['is_system'],
        unique=False
    )

    op.create_index(
        'action_definitions_v2_project_id',
        'action_definitions_v2',
        ['project_id'],
        unique=False
    )

    op.create_index(
        'action_definitions_v2_scope',
        'action_definitions_v2',
        ['scope'],
        unique=False
    )

    op.create_index(
        'cron_triggers_v2_next_execution_time',
        'cron_triggers_v2',
        ['next_execution_time'],
        unique=False
    )

    op.create_index(
        'cron_triggers_v2_project_id',
        'cron_triggers_v2',
        ['project_id'],
        unique=False
    )

    op.create_index(
        'cron_triggers_v2_scope',
        'cron_triggers_v2',
        ['scope'],
        unique=False
    )

    op.create_index(
        'cron_triggers_v2_workflow_name',
        'cron_triggers_v2',
        ['workflow_name'],
        unique=False
    )

    cron_v2_constrs = [uc['name'] for uc in
                       inspector.get_unique_constraints('cron_triggers_v2')]
    if ('cron_triggers_v2_workflow_input_hash_workflow_name_pattern__key' in
            cron_v2_constrs):
        op.drop_constraint(
            'cron_triggers_v2_workflow_input_hash_workflow_name_pattern__key',
            'cron_triggers_v2',
            type_='unique'
        )

    if ('cron_triggers_v2_workflow_input_hash_workflow_name_pattern_key1' in
            cron_v2_constrs):
        op.drop_constraint(
            'cron_triggers_v2_workflow_input_hash_workflow_name_pattern_key1',
            'cron_triggers_v2',
            type_='unique'
        )

    op.create_index(
        'delayed_calls_v2_processing_execution_time',
        'delayed_calls_v2',
        ['processing', 'execution_time'],
        unique=False
    )

    op.create_index(
        'environments_v2_name',
        'environments_v2',
        ['name'],
        unique=False
    )

    op.create_index(
        'environments_v2_project_id',
        'environments_v2',
        ['project_id'],
        unique=False
    )

    op.create_index(
        'environments_v2_scope',
        'environments_v2',
        ['scope'],
        unique=False
    )

    op.create_index(
        'executions_v2_project_id',
        'executions_v2',
        ['project_id'],
        unique=False
    )

    op.create_index(
        'executions_v2_scope',
        'executions_v2',
        ['scope'],
        unique=False
    )

    op.create_index(
        'executions_v2_state',
        'executions_v2',
        ['state'],
        unique=False
    )

    op.create_index(
        'executions_v2_task_execution_id',
        'executions_v2',
        ['task_execution_id'],
        unique=False
    )

    op.create_index(
        'executions_v2_type',
        'executions_v2',
        ['type'],
        unique=False
    )

    op.create_index(
        'executions_v2_updated_at',
        'executions_v2',
        ['updated_at'],
        unique=False
    )

    op.create_index(
        'executions_v2_workflow_execution_id',
        'executions_v2',
        ['workflow_execution_id'],
        unique=False
    )

    op.create_index(
        'workbooks_v2_project_id',
        'workbooks_v2',
        ['project_id'],
        unique=False
    )

    op.create_index(
        'workbooks_v2_scope',
        'workbooks_v2',
        ['scope'],
        unique=False
    )

    op.create_index(
        'workflow_definitions_v2_is_system',
        'workflow_definitions_v2',
        ['is_system'],
        unique=False
    )

    op.create_index(
        'workflow_definitions_v2_project_id',
        'workflow_definitions_v2',
        ['project_id'],
        unique=False
    )

    op.create_index(
        'workflow_definitions_v2_scope',
        'workflow_definitions_v2',
        ['scope'],
        unique=False
    )



# Copyright 2013 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
# Copyright 2016 - Brocade Communications Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


# TODO(rakhmerov): Can we make one parent for errors and exceptions?

class MistralError(Exception):
    """Mistral specific error.

    Reserved for situations that can't be automatically handled. When it occurs
    it signals that there is a major environmental problem like invalid startup
    configuration or implementation problem (e.g. some code doesn't take care
    of certain corner cases). From architectural perspective it's pointless to
    try to handle this type of problems except doing some finalization work
    like transaction rollback, deleting temporary files etc.
    """

    message = "An unknown error occurred"
    http_code = 500

    def __init__(self, message=None):
        if message is not None:
            self.message = message

        super(MistralError, self).__init__(
            '%d: %s' % (self.http_code, self.message))

    @property
    def code(self):
        """This is here for webob to read.

        https://github.com/Pylons/webob/blob/master/webob/exc.py
        """
        return self.http_code

    def __str__(self):
        return self.message


class MistralException(Exception):
    """Mistral specific exception.

    Reserved for situations that are not critical for program continuation.
    It is possible to recover from this type of problems automatically and
    continue program execution. Such problems may be related with invalid user
    input (such as invalid syntax) or temporary environmental problems.

    In case if an instance of a certain exception type bubbles up to API layer
    then this type of exception it must be associated with an http code so it's
    clear how to represent it for a client.

    To correctly use this class, inherit from it and define a 'message' and
    'http_code' properties.
    """
    message = "An unknown exception occurred"
    http_code = 500

    def __init__(self, message=None):
        if message is not None:
            self.message = message

        super(MistralException, self).__init__(
            '%d: %s' % (self.http_code, self.message))

    @property
    def code(self):
        """This is here for webob to read.

        https://github.com/Pylons/webob/blob/master/webob/exc.py
        """
        return self.http_code

    def __str__(self):
        return self.message


# Database errors.

class DBError(MistralError):
    http_code = 400


class DBDuplicateEntryError(DBError):
    http_code = 409
    message = "Database object already exists"


class DBQueryEntryError(DBError):
    http_code = 400


class DBEntityNotFoundError(DBError):
    http_code = 404
    message = "Object not found"


# DSL exceptions.

class DSLParsingException(MistralException):
    http_code = 400


class ExpressionGrammarException(DSLParsingException):
    http_code = 400


class JinjaGrammarException(ExpressionGrammarException):
    message = "Invalid grammar of Jinja expression"


class YaqlGrammarException(ExpressionGrammarException):
    message = "Invalid grammar of YAQL expression"


class InvalidModelException(DSLParsingException):
    http_code = 400
    message = "Wrong entity definition"


# Various common exceptions and errors.

class EvaluationException(MistralException):
    http_code = 400


class JinjaEvaluationException(EvaluationException):
    message = "Can not evaluate Jinja expression"


class YaqlEvaluationException(EvaluationException):
    message = "Can not evaluate YAQL expression"


class DataAccessException(MistralException):
    http_code = 400


class ActionException(MistralException):
    http_code = 400


class InvalidActionException(MistralException):
    http_code = 400


class ActionRegistrationException(MistralException):
    message = "Failed to register action"


class EngineException(MistralException):
    http_code = 500


class WorkflowException(MistralException):
    http_code = 400


class EventTriggerException(MistralException):
    http_code = 400


class InputException(MistralException):
    http_code = 400


class ApplicationContextNotFoundException(MistralException):
    http_code = 400
    message = "Application context not found"


class InvalidResultException(MistralException):
    http_code = 400
    message = "Unable to parse result"


class SizeLimitExceededException(MistralException):
    http_code = 400

    def __init__(self, field_name, size_kb, size_limit_kb):
        super(SizeLimitExceededException, self).__init__(
            "Size of '%s' is %dKB which exceeds the limit of %dKB"
            % (field_name, size_kb, size_limit_kb))


class CoordinationException(MistralException):
    http_code = 500


class NotAllowedException(MistralException):
    http_code = 403
    message = "Operation not allowed"


class UnauthorizedException(MistralException):
    http_code = 401
    message = "Unauthorized"


class KombuException(Exception):
    def __init__(self, e):
        super(KombuException, self).__init__(e)

        self.exc_type = e.__class__.__name__
        self.value = str(e)



# Copyright 2014 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral.db.v2 import api as db_api
from mistral.tests.unit import base


class ActionManagerTest(base.DbTestCase):
    def test_action_input(self):
        std_http = db_api.get_action_definition("std.http")
        std_email = db_api.get_action_definition("std.email")

        http_action_input = (
            'url, method="GET", params=null, body=null, '
            'headers=null, cookies=null, auth=null, '
            'timeout=null, allow_redirects=null, '
            'proxies=null, verify=null'
        )

        self.assertEqual(http_action_input, std_http.input)

        std_email_input = (
            "from_addr, to_addrs, smtp_server, "
            "smtp_password=null, subject=null, body=null"
        )

        self.assertEqual(std_email_input, std_email.input)

    def test_action_description(self):
        std_http = db_api.get_action_definition("std.http")
        std_echo = db_api.get_action_definition("std.echo")

        self.assertIn("Constructs an HTTP action", std_http.description)
        self.assertIn("param body: (optional) Dictionary, bytes",
                      std_http.description)

        self.assertIn("This action just returns a configured value",
                      std_echo.description)



# Copyright 2015 - Alcatel-lucent, Inc.
# Copyright 2015 - StackStorm, Inc.
# Copyright 2016 - Brocade Communications Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import datetime

from mistral import context as ctx
from mistral.db.v2 import api as db_api
from mistral.services import expiration_policy
from mistral.tests.unit import base
from oslo_config import cfg


def _create_workflow_executions():
    time_now = datetime.datetime.now()

    wf_execs = [
        {
            'id': '123',
            'name': 'success_expired',
            'created_at': time_now - datetime.timedelta(minutes=60),
            'updated_at': time_now - datetime.timedelta(minutes=59),
            'workflow_name': 'test_exec',
            'state': "SUCCESS",
        },
        {
            'id': '456',
            'name': 'error_expired',
            'created_at': time_now - datetime.timedelta(days=3, minutes=10),
            'updated_at': time_now - datetime.timedelta(days=3),
            'workflow_name': 'test_exec',
            'state': "ERROR",
        },
        {
            'id': '789',
            'name': 'running_not_expired',
            'created_at': time_now - datetime.timedelta(days=3, minutes=10),
            'updated_at': time_now - datetime.timedelta(days=3),
            'workflow_name': 'test_exec',
            'state': "RUNNING",
        },
        {
            'id': '987',
            'name': 'success_not_expired',
            'created_at': time_now - datetime.timedelta(minutes=15),
            'updated_at': time_now - datetime.timedelta(minutes=5),
            'workflow_name': 'test_exec',
            'state': "SUCCESS",
        },
        {
            'id': 'abc',
            'name': 'cancelled_expired',
            'created_at': time_now - datetime.timedelta(minutes=60),
            'updated_at': time_now - datetime.timedelta(minutes=59),
            'workflow_name': 'test_exec',
            'state': "CANCELLED",
        },
        {
            'id': 'def',
            'name': 'cancelled_not_expired',
            'created_at': time_now - datetime.timedelta(minutes=15),
            'updated_at': time_now - datetime.timedelta(minutes=5),
            'workflow_name': 'test_exec',
            'state': "CANCELLED",
        }
    ]

    for wf_exec in wf_execs:
        db_api.create_workflow_execution(wf_exec)

    # Create a nested workflow execution.

    db_api.create_task_execution(
        {
            'id': '789',
            'workflow_execution_id': '987',
            'name': 'my_task'
        }
    )

    db_api.create_workflow_execution(
        {
            'id': '654',
            'name': 'expired but not a parent',
            'created_at': time_now - datetime.timedelta(days=15),
            'updated_at': time_now - datetime.timedelta(days=10),
            'workflow_name': 'test_exec',
            'state': "SUCCESS",
            'task_execution_id': '789'
        }
    )


def _switch_context(project_id, is_admin):
    _ctx = ctx.MistralContext(
        user_id=None,
        project_id=project_id,
        auth_token=None,
        is_admin=is_admin
    )

    ctx.set_ctx(_ctx)


class ExpirationPolicyTest(base.DbTestCase):
    def test_expiration_policy_for_executions(self):
        # Delete execution uses a secured filtering and we need
        # to verify that admin able to do that for other projects.
        cfg.CONF.set_default('auth_enable', True, group='pecan')

        # Since we are removing other projects execution,
        # we want to load the executions with other project_id.
        _switch_context('non_admin_project', False)

        _create_workflow_executions()

        now = datetime.datetime.now()

        # This execution has a parent wf and testing that we are
        # querying only for parent wfs.
        exec_child = db_api.get_workflow_execution('654')

        self.assertEqual('789', exec_child.task_execution_id)

        # Call for all expired wfs execs.
        execs = db_api.get_expired_executions(now)

        # Should be only 5, the RUNNING execution shouldn't return,
        # so the child wf (that has parent task id).
        self.assertEqual(5, len(execs))

        # Switch context to Admin since expiration policy running as Admin.
        _switch_context(None, True)

        _set_expiration_policy_config(1, 30)
        expiration_policy.run_execution_expiration_policy(self, ctx)

        # Only non_expired available (update_at < older_than).
        execs = db_api.get_expired_executions(now)

        self.assertEqual(2, len(execs))
        self.assertListEqual(['987', 'def'], sorted([ex.id for ex in execs]))

        _set_expiration_policy_config(1, 5)
        expiration_policy.run_execution_expiration_policy(self, ctx)
        execs = db_api.get_expired_executions(now)

        self.assertEqual(0, len(execs))

    def test_negative_wrong_conf_values(self):
        _set_expiration_policy_config(None, None)
        e_policy = expiration_policy.ExecutionExpirationPolicy(cfg.CONF)

        self.assertDictEqual({}, e_policy._periodic_spacing)
        self.assertListEqual([], e_policy._periodic_tasks)

        _set_expiration_policy_config(None, 60)
        e_policy = expiration_policy.ExecutionExpirationPolicy(cfg.CONF)

        self.assertDictEqual({}, e_policy._periodic_spacing)
        self.assertListEqual([], e_policy._periodic_tasks)

        _set_expiration_policy_config(60, None)
        e_policy = expiration_policy.ExecutionExpirationPolicy(cfg.CONF)

        self.assertDictEqual({}, e_policy._periodic_spacing)
        self.assertListEqual([], e_policy._periodic_tasks)

    def test_periodic_task_parameters(self):
        _set_expiration_policy_config(17, 13)

        e_policy = expiration_policy.ExecutionExpirationPolicy(cfg.CONF)

        self.assertEqual(17 * 60, e_policy._periodic_spacing
                         ['run_execution_expiration_policy'])

    def tearDown(self):
        """Restores the size limit config to default."""
        super(ExpirationPolicyTest, self).tearDown()

        cfg.CONF.set_default('auth_enable', False, group='pecan')

        ctx.set_ctx(None)

        _set_expiration_policy_config(None, None)


def _set_expiration_policy_config(evaluation_interval, older_than):
    cfg.CONF.set_default(
        'evaluation_interval',
        evaluation_interval,
        group='execution_expiration_policy'
    )
    cfg.CONF.set_default(
        'older_than',
        older_than,
        group='execution_expiration_policy'
    )



# Copyright 2014 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import mock
from oslo_config import cfg

from mistral.actions import std_actions
from mistral import context as auth_context
from mistral.db.v2 import api as db_api
from mistral import exceptions as exc
from mistral.services import workbooks as wb_service
from mistral.tests.unit.engine import base
from mistral.workflow import states


# Use the set_default method to set value otherwise in certain test cases
# the change in value is not permanent.
cfg.CONF.set_default('auth_enable', False, group='pecan')


WB1 = """
---
version: '2.0'

name: wb1

workflows:
  wf1:
    type: reverse

    input:
      - param1
      - param2
    output:
      final_result: <% $.final_result %>

    tasks:
      task1:
        action: std.echo output=<% $.param1 %>
        publish:
          result1: <% task(task1).result %>

      task2:
        action: std.echo output="'<% $.param1 %> & <% $.param2 %>'"
        publish:
          final_result: <% task(task2).result %>
        requires: [task1]

  wf2:
    type: direct

    output:
      slogan: <% $.slogan %>

    tasks:
      task1:
        workflow: wf1 param1='Bonnie' param2='Clyde' task_name='task2'
        publish:
          slogan: "<% task(task1).result.final_result %> is a cool movie!"
"""

WB2 = """
---
version: '2.0'

name: wb2

workflows:
  wf1:
    type: direct

    tasks:
      task1:
        workflow: wf2

  wf2:
    type: direct

    output:
      var1: <% $.does_not_exist %>

    tasks:
        task1:
            action: std.noop
"""


class SubworkflowsTest(base.EngineTestCase):
    def setUp(self):
        super(SubworkflowsTest, self).setUp()

        wb_service.create_workbook_v2(WB1)
        wb_service.create_workbook_v2(WB2)

    def test_subworkflow_success(self):
        wf2_ex = self.engine.start_workflow('wb1.wf2', None)

        project_id = auth_context.ctx().project_id

        # Execution of 'wf2'.
        self.assertEqual(project_id, wf2_ex.project_id)
        self.assertIsNotNone(wf2_ex)
        self.assertDictEqual({}, wf2_ex.input)
        self.assertDictEqual({}, wf2_ex.params)

        self._await(lambda: len(db_api.get_workflow_executions()) == 2, 0.5, 5)

        wf_execs = db_api.get_workflow_executions()

        self.assertEqual(2, len(wf_execs))

        # Execution of 'wf2'.
        wf1_ex = self._assert_single_item(wf_execs, name='wb1.wf1')
        wf2_ex = self._assert_single_item(wf_execs, name='wb1.wf2')

        self.assertEqual(project_id, wf1_ex.project_id)
        self.assertIsNotNone(wf1_ex.task_execution_id)
        self.assertDictContainsSubset(
            {
                'task_name': 'task2',
                'task_execution_id': wf1_ex.task_execution_id
            },
            wf1_ex.params
        )
        self.assertDictEqual(
            {
                'param1': 'Bonnie',
                'param2': 'Clyde'
            },
            wf1_ex.input
        )

        # Wait till workflow 'wf1' is completed.
        self.await_workflow_success(wf1_ex.id)

        with db_api.transaction():
            wf1_ex = db_api.get_workflow_execution(wf1_ex.id)

            wf1_output = wf1_ex.output

        self.assertDictEqual(
            {'final_result': "'Bonnie & Clyde'"},
            wf1_output
        )

        # Wait till workflow 'wf2' is completed.
        self.await_workflow_success(wf2_ex.id, timeout=4)

        with db_api.transaction():
            wf2_ex = db_api.get_workflow_execution(wf2_ex.id)

            wf2_output = wf2_ex.output

        self.assertDictEqual(
            {'slogan': "'Bonnie & Clyde' is a cool movie!"},
            wf2_output
        )

        # Check project_id in tasks.
        wf1_task_execs = db_api.get_task_executions(
            workflow_execution_id=wf1_ex.id
        )
        wf2_task_execs = db_api.get_task_executions(
            workflow_execution_id=wf2_ex.id
        )

        wf2_task1_ex = self._assert_single_item(wf1_task_execs, name='task1')
        wf1_task1_ex = self._assert_single_item(wf2_task_execs, name='task1')
        wf1_task2_ex = self._assert_single_item(wf1_task_execs, name='task2')

        self.assertEqual(project_id, wf2_task1_ex.project_id)
        self.assertEqual(project_id, wf1_task1_ex.project_id)
        self.assertEqual(project_id, wf1_task2_ex.project_id)

    @mock.patch.object(std_actions.EchoAction, 'run',
                       mock.MagicMock(side_effect=exc.ActionException))
    def test_subworkflow_error(self):
        self.engine.start_workflow('wb1.wf2', None)

        self._await(lambda: len(db_api.get_workflow_executions()) == 2, 0.5, 5)

        wf_execs = db_api.get_workflow_executions()

        self.assertEqual(2, len(wf_execs))

        wf1_ex = self._assert_single_item(wf_execs, name='wb1.wf1')
        wf2_ex = self._assert_single_item(wf_execs, name='wb1.wf2')

        # Wait till workflow 'wf1' is completed.
        self.await_workflow_error(wf1_ex.id)

        # Wait till workflow 'wf2' is completed, its state must be ERROR.
        self.await_workflow_error(wf2_ex.id)

    def test_subworkflow_yaql_error(self):
        wf_ex = self.engine.start_workflow('wb2.wf1', None)

        self.await_workflow_error(wf_ex.id)

        wf_execs = db_api.get_workflow_executions()

        self.assertEqual(2, len(wf_execs))

        wf2_ex = self._assert_single_item(wf_execs, name='wb2.wf2')

        self.assertEqual(states.ERROR, wf2_ex.state)
        self.assertIn('Can not evaluate YAQL expression', wf2_ex.state_info)

        # Ensure error message is bubbled up to the main workflow.
        wf1_ex = self._assert_single_item(wf_execs, name='wb2.wf1')

        self.assertEqual(states.ERROR, wf1_ex.state)
        self.assertIn('Can not evaluate YAQL expression', wf1_ex.state_info)

    def test_subworkflow_environment_inheritance(self):
        env = {'key1': 'abc'}

        wf2_ex = self.engine.start_workflow('wb1.wf2', None, env=env)

        # Execution of 'wf2'.
        self.assertIsNotNone(wf2_ex)
        self.assertDictEqual({}, wf2_ex.input)
        self.assertDictEqual({'env': env}, wf2_ex.params)

        self._await(lambda: len(db_api.get_workflow_executions()) == 2, 0.5, 5)

        wf_execs = db_api.get_workflow_executions()

        self.assertEqual(2, len(wf_execs))

        # Execution of 'wf1'.
        wf1_ex = self._assert_single_item(wf_execs, name='wb1.wf1')
        wf2_ex = self._assert_single_item(wf_execs, name='wb1.wf2')

        expected_start_params = {
            'task_name': 'task2',
            'task_execution_id': wf1_ex.task_execution_id,
            'env': env
        }

        self.assertIsNotNone(wf1_ex.task_execution_id)
        self.assertDictContainsSubset(expected_start_params, wf1_ex.params)

        # Wait till workflow 'wf1' is completed.
        self.await_workflow_success(wf1_ex.id)

        # Wait till workflow 'wf2' is completed.
        self.await_workflow_success(wf2_ex.id)



# Copyright 2015 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from oslo_config import cfg

from mistral.db.v2.sqlalchemy import api as db_api
from mistral.db.v2.sqlalchemy import models as db_models
from mistral.tests.unit import base as test_base

WF_EX = {
    'id': '1',
    'spec': {},
    'start_params': {'task': 'my_task1'},
    'state': 'IDLE',
    'state_info': "Running...",
    'created_at': None,
    'updated_at': None,
    'context': None,
    'task_id': None,
    'trust_id': None,
    'description': None,
    'output': None
}


TASK_EX = {
    'workflow_execution_id': '1',
    'workflow_name': 'my_wf',
    'name': 'my_task1',
    'spec': None,
    'action_spec': None,
    'state': 'IDLE',
    'tags': ['deployment'],
    'in_context': None,
    'runtime_context': None,
    'created_at': None,
    'updated_at': None
}


class InsertOrIgnoreTest(test_base.DbTestCase):
    def setUp(self):
        super(InsertOrIgnoreTest, self).setUp()

        cfg.CONF.set_default('auth_enable', True, group='pecan')

        db_api.create_workflow_execution(WF_EX)

        self.addCleanup(
            cfg.CONF.set_default,
            'auth_enable',
            False,
            group='pecan'
        )

    def test_insert_or_ignore_without_conflicts(self):
        db_api.insert_or_ignore(
            db_models.TaskExecution,
            TASK_EX.copy()
        )

        task_execs = db_api.get_task_executions()

        self.assertEqual(1, len(task_execs))

        task_ex = task_execs[0]

        self._assert_dict_contains_subset(TASK_EX, task_ex.to_dict())

    def test_insert_or_ignore_with_conflicts(self):
        # Insert the first object.
        values = TASK_EX.copy()

        values['unique_key'] = 'key'

        db_api.insert_or_ignore(db_models.TaskExecution, values)

        task_execs = db_api.get_task_executions()

        self.assertEqual(1, len(task_execs))

        task_ex = task_execs[0]

        self._assert_dict_contains_subset(TASK_EX, task_ex.to_dict())

        # Insert the second object with the same unique key.
        # We must not get exceptions and new object must not be saved.
        values = TASK_EX.copy()

        values['unique_key'] = 'key'

        db_api.insert_or_ignore(db_models.TaskExecution, values)

        task_execs = db_api.get_task_executions()

        self.assertEqual(1, len(task_execs))

        task_ex = task_execs[0]

        self._assert_dict_contains_subset(TASK_EX, task_ex.to_dict())



# Copyright 2015 - Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_config import cfg

from mistral.db.v2 import api as db_api
from mistral.services import workflows as wf_service
from mistral.tests.unit.engine import base as engine_test_base
from mistral.workflow import states


# Use the set_default method to set value otherwise in certain test cases
# the change in value is not permanent.
cfg.CONF.set_default('auth_enable', False, group='pecan')


class YAQLFunctionsEngineTest(engine_test_base.EngineTestCase):
    def test_task_function(self):
        wf_text = """---
        version: '2.0'

        wf:
          type: direct

          tasks:
            task1:
              description: This is task 1
              tags: ['t1']
              action: std.echo output=1
              publish:
                name: <% task(task1).name %>
                description: <% task(task1).spec.description %>
                tags: <% task(task1).spec.tags%>
                state: <% task(task1).state %>
                state_info: <% task(task1).state_info %>
                res: <% task(task1).result %>
              on-success:
                - task2

            task2:
              action: std.echo output=<% task(task1).result + 1 %>
              publish:
                name: <% task(task1).name %>
                description: <% task(task1).spec.description %>
                tags: <% task(task1).spec.tags%>
                state: <% task(task1).state %>
                state_info: <% task(task1).state_info %>
                res: <% task(task1).result %>
                task2_res: <% task(task2).result %>
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf', {})

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            # Reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            tasks = wf_ex.task_executions

        self.assertEqual(states.SUCCESS, wf_ex.state)

        task1 = self._assert_single_item(
            tasks,
            name='task1',
            state=states.SUCCESS
        )
        task2 = self._assert_single_item(
            tasks,
            name='task2',
            state=states.SUCCESS
        )

        self.assertDictEqual(
            {
                'name': 'task1',
                'description': 'This is task 1',
                'tags': ['t1'],
                'state': states.SUCCESS,
                'state_info': None,
                'res': 1
            },
            task1.published
        )
        self.assertDictEqual(
            {
                'name': 'task1',
                'description': 'This is task 1',
                'tags': ['t1'],
                'state': states.SUCCESS,
                'state_info': None,
                'res': 1,
                'task2_res': 2
            },
            task2.published
        )

    def test_task_function_returns_null(self):
        wf_text = """---
            version: '2.0'

            wf:
              output:
                task2: <% task(task2) %>
                task2bool: <% task(task2) = null %>

              tasks:
                task1:
                  action: std.noop
                  on-success:
                    - task2: <% false %>

                task2:
                  action: std.noop
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf', {})

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            self.assertDictEqual(
                {
                    'task2': None,
                    'task2bool': True
                },
                wf_ex.output
            )

            task_execs = wf_ex.task_executions

        self.assertEqual(1, len(task_execs))

    def test_task_function_non_existing(self):
        wf_text = """---
            version: '2.0'

            wf:
              type: direct
              output:
                task_name: <% task(non_existing_task).name %>

              tasks:
                task1:
                  action: std.noop
            """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf', {})

        self.await_workflow_error(wf_ex.id)

        wf_ex = db_api.get_workflow_execution(wf_ex.id)

        self.assertEqual(states.ERROR, wf_ex.state)
        self.assertIn('non_existing_task', wf_ex.state_info)

    def test_task_function_no_arguments(self):
        wf_text = """---
            version: '2.0'

            wf:
              tasks:
                task1:
                  action: std.echo output=1
                  publish:
                    task1_id: <% task().id %>
                    task1_result:  <% task().result %>
                    task1_state: <% task().state %>
                  on-success: task2

                task2:
                  action: std.echo output=2
                  publish:
                    task2_id: <% task().id %>
                    task2_result:  <% task().result %>
                    task2_state: <% task().state %>
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf', {})

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            task1_ex = self._assert_single_item(
                wf_ex.task_executions, name='task1'
            )
            task2_ex = self._assert_single_item(
                wf_ex.task_executions, name='task2'
            )

            self.assertDictEqual(
                {
                    'task1_id': task1_ex.id,
                    'task1_result': 1,
                    'task1_state': states.SUCCESS
                },
                task1_ex.published
            )
            self.assertDictEqual(
                {
                    'task2_id': task2_ex.id,
                    'task2_result': 2,
                    'task2_state': states.SUCCESS
                },
                task2_ex.published
            )

    def test_uuid_function(self):
        wf_text = """---
            version: '2.0'

            wf:
              tasks:
                task1:
                  action: std.echo output=<% uuid() %>
                  publish:
                    result: <% task(task1).result %>
            """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf', {})

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            task_execs = wf_ex.task_executions

        task_ex = task_execs[0]

        result = task_ex.published['result']

        self.assertIsNotNone(result)
        self.assertEqual(36, len(result))
        self.assertEqual(4, result.count('-'))

    def test_execution_function(self):
        wf_text = """---
            version: '2.0'

            wf:
              input:
                - k1
                - k2: v2_default

              tasks:
                task1:
                  action: std.echo output=<% execution() %>
                  publish:
                    result: <% task(task1).result %>
            """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow(
            'wf',
            {'k1': 'v1'},
            param1='blablabla'
        )

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            task_execs = wf_ex.task_executions

        task_ex = task_execs[0]

        execution = task_ex.published['result']

        self.assertIsInstance(execution, dict)

        spec = execution['spec']

        self.assertEqual('2.0', spec['version'])
        self.assertEqual('wf', spec['name'])
        self.assertIn('tasks', spec)
        self.assertEqual(1, len(spec['tasks']))

        self.assertDictEqual(
            {
                'k1': 'v1',
                'k2': 'v2_default'
            },
            execution['input']
        )

        self.assertDictEqual({'param1': 'blablabla'}, execution['params'])



# Copyright 2015 - Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import amqp
import socket
import threading
import time

import kombu
from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging as messaging
from stevedore import driver

from mistral import context as auth_ctx
from mistral.engine.rpc_backend import base as rpc_base
from mistral.engine.rpc_backend.kombu import base as kombu_base
from mistral.engine.rpc_backend.kombu import kombu_hosts
from mistral import exceptions as exc


LOG = logging.getLogger(__name__)
CONF = cfg.CONF
_pool_opts = [
    cfg.IntOpt('executor_thread_pool_size',
               default=64,
               deprecated_name="rpc_thread_pool_size",
               help='Size of executor thread pool.'),
]


class KombuRPCServer(rpc_base.RPCServer, kombu_base.Base):
    def __init__(self, conf):
        super(KombuRPCServer, self).__init__(conf)

        self._register_mistral_serialization()
        CONF.register_opts(_pool_opts)

        self._transport_url = messaging.TransportURL.parse(
            CONF,
            CONF.transport_url
        )
        self._check_backend()

        self.topic = conf.topic
        self.server_id = conf.host

        self._hosts = kombu_hosts.KombuHosts(CONF)

        self._executor_threads = CONF.executor_thread_pool_size
        self.exchange = CONF.control_exchange
        self.virtual_host = CONF.oslo_messaging_rabbit.rabbit_virtual_host
        self.durable_queue = CONF.oslo_messaging_rabbit.amqp_durable_queues
        self.auto_delete = CONF.oslo_messaging_rabbit.amqp_auto_delete
        self.routing_key = self.topic
        self.channel = None
        self.conn = None
        self._running = threading.Event()
        self._stopped = threading.Event()
        self.endpoints = []
        self._worker = None

        # TODO(ddeja): Those 2 options should be gathered from config.
        self._sleep_time = 1
        self._max_sleep_time = 512

    @property
    def is_running(self):
        """Return whether server is running."""
        return self._running.is_set()

    def run(self, executor='blocking'):
        """Start the server."""
        self._prepare_worker(executor)

        while True:
            try:
                host = self._hosts.get_host()

                self.conn = self._make_connection(
                    host.hostname,
                    host.port,
                    host.username,
                    host.password,
                    self.virtual_host,
                )

                conn = kombu.connections[self.conn].acquire(block=True)

                exchange = self._make_exchange(
                    self.exchange,
                    durable=self.durable_queue,
                    auto_delete=self.auto_delete
                )

                queue = self._make_queue(
                    self.topic,
                    exchange,
                    routing_key=self.routing_key,
                    durable=self.durable_queue,
                    auto_delete=self.auto_delete
                )
                with conn.Consumer(
                        queues=queue,
                        callbacks=[self._process_message],
                ) as consumer:
                    consumer.qos(prefetch_count=1)

                    self._running.set()
                    self._stopped.clear()

                    LOG.info("Connected to AMQP at %s:%s" % (
                        host.hostname,
                        host.port
                    ))

                    while self.is_running:
                        try:
                            conn.drain_events(timeout=1)
                        except socket.timeout:
                            pass
                        except KeyboardInterrupt:
                            self.stop()

                            LOG.info("Server with id='{0}' stopped.".format(
                                self.server_id))

                            return
            except (socket.error, amqp.exceptions.ConnectionForced) as e:
                LOG.debug("Broker connection failed: %s" % e)
            finally:
                self._stopped.set()

                LOG.debug("Sleeping for %s seconds, than retrying connection" %
                          self._sleep_time
                          )

                time.sleep(self._sleep_time)

                self._sleep_time = min(
                    self._sleep_time * 2,
                    self._max_sleep_time
                )

    def stop(self, graceful=False):
        self._running.clear()

        if graceful:
            self.wait()

    def wait(self):
        self._stopped.wait()
        try:
            self._worker.shutdown(wait=True)
        except AttributeError as e:
            LOG.warning("Cannot stop worker in graceful way: %s" % e)

    def _get_rpc_method(self, method_name):
        for endpoint in self.endpoints:
            if hasattr(endpoint, method_name):
                return getattr(endpoint, method_name)

        return None

    @staticmethod
    def _set_auth_ctx(ctx):
        if not isinstance(ctx, dict):
            return

        context = auth_ctx.MistralContext(**ctx)
        auth_ctx.set_ctx(context)

        return context

    def publish_message(self, body, reply_to, corr_id, res_type='response'):
        with kombu.producers[self.conn].acquire(block=True) as producer:
            producer.publish(
                body=body,
                exchange=self.exchange,
                routing_key=reply_to,
                correlation_id=corr_id,
                serializer='pickle' if res_type == 'error'
                else 'mistral_serialization',
                type=res_type
            )

    def _on_message_safe(self, request, message):
        try:
            return self._on_message(request, message)
        except Exception as e:
            # Wrap exception into another exception for compability with oslo.
            self.publish_message(
                exc.KombuException(e),
                message.properties['reply_to'],
                message.properties['correlation_id'],
                res_type='error'
            )
        finally:
            message.ack()

    def _on_message(self, request, message):
        LOG.debug('Received message %s',
                  request)

        is_async = request.get('async', False)
        rpc_ctx = request.get('rpc_ctx')
        redelivered = message.delivery_info.get('redelivered', None)
        rpc_method_name = request.get('rpc_method')
        arguments = request.get('arguments')
        correlation_id = message.properties['correlation_id']
        reply_to = message.properties['reply_to']

        if redelivered is not None:
            rpc_ctx['redelivered'] = redelivered

        rpc_context = self._set_auth_ctx(rpc_ctx)

        rpc_method = self._get_rpc_method(rpc_method_name)

        if not rpc_method:
            raise exc.MistralException("No such method: %s" % rpc_method_name)

        response = rpc_method(rpc_ctx=rpc_context, **arguments)

        if not is_async:
            self.publish_message(
                response,
                reply_to,
                correlation_id
            )

    def register_endpoint(self, endpoint):
        self.endpoints.append(endpoint)

    def _process_message(self, request, message):
        self._worker.submit(self._on_message_safe, request, message)

    def _prepare_worker(self, executor='blocking'):
        mgr = driver.DriverManager('kombu_driver.executors', executor)

        executor_opts = {}
        if executor == 'threading':
            executor_opts['max_workers'] = self._executor_threads

        self._worker = mgr.driver(**executor_opts)



# Copyright 2014 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from mistral.actions import std_actions as std
from mistral.services import action_manager as a_m
from mistral.tests.unit import base


class ActionManagerTest(base.DbTestCase):

    def test_register_standard_actions(self):
        action_list = a_m.get_registered_actions()

        self._assert_single_item(action_list, name="std.echo")
        self._assert_single_item(action_list, name="std.email")
        self._assert_single_item(action_list, name="std.http")
        self._assert_single_item(action_list, name="std.mistral_http")
        self._assert_single_item(action_list, name="std.ssh")
        self._assert_single_item(action_list, name="std.javascript")

        self._assert_single_item(action_list, name="nova.servers_get")
        self._assert_single_item(
            action_list,
            name="nova.volumes_delete_server_volume"
        )

        server_find_action = self._assert_single_item(
            action_list,
            name="nova.servers_find"
        )
        self.assertIn('**', server_find_action.input)

        self._assert_single_item(action_list, name="keystone.users_list")
        self._assert_single_item(action_list, name="keystone.trusts_create")

        self._assert_single_item(action_list, name="glance.images_list")
        self._assert_single_item(action_list, name="glance.images_delete")

    def test_get_action_class(self):
        self.assertTrue(
            issubclass(a_m.get_action_class("std.echo"), std.EchoAction)
        )
        self.assertTrue(
            issubclass(a_m.get_action_class("std.http"), std.HTTPAction)
        )
        self.assertTrue(
            issubclass(
                a_m.get_action_class("std.mistral_http"),
                std.MistralHTTPAction
            )
        )
        self.assertTrue(
            issubclass(a_m.get_action_class("std.email"), std.SendEmailAction)
        )
        self.assertTrue(
            issubclass(
                a_m.get_action_class("std.javascript"),
                std.JavaScriptAction
            )
        )



# Copyright 2014 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import copy
import mock

from mistral.db.v2 import api as db_api
from mistral.db.v2.sqlalchemy import models
from mistral import exceptions as exc
from mistral.tests.unit.api import base
from mistral import utils


ACTION_DEFINITION = """
---
version: '2.0'

my_action:
  description: My super cool action.
  tags: ['test', 'v2']
  base: std.echo
  base-input:
    output: "{$.str1}{$.str2}"
"""

ACTION_DEFINITION_INVALID_NO_BASE = """
---
version: '2.0'

my_action:
  description: My super cool action.
  tags: ['test', 'v2']

  base-input:
    output: "{$.str1}{$.str2}"
"""

ACTION_DEFINITION_INVALID_YAQL = """
---
version: '2.0'

my_action:
  description: My super cool action.
  tags: ['test', 'v2']
  base: std.echo
  base-input:
    output: <% $. %>
"""

ACTION_DSL_PARSE_EXCEPTION = """
---
%
"""

SYSTEM_ACTION_DEFINITION = """
---
version: '2.0'

std.echo:
  base: std.http
  base-input:
    url: "some.url"
"""

ACTION = {
    'id': '123e4567-e89b-12d3-a456-426655440000',
    'name': 'my_action',
    'is_system': False,
    'description': 'My super cool action.',
    'tags': ['test', 'v2'],
    'definition': ACTION_DEFINITION
}

SYSTEM_ACTION = {
    'id': '1234',
    'name': 'std.echo',
    'is_system': True,
    'definition': SYSTEM_ACTION_DEFINITION
}

ACTION_DB = models.ActionDefinition()
ACTION_DB.update(ACTION)

SYSTEM_ACTION_DB = models.ActionDefinition()
SYSTEM_ACTION_DB.update(SYSTEM_ACTION)

UPDATED_ACTION_DEFINITION = """
---
version: '2.0'

my_action:
  description: My super cool action.
  base: std.echo
  base-input:
    output: "{$.str1}{$.str2}{$.str3}"
"""

UPDATED_ACTION_DB = copy.copy(ACTION_DB)
UPDATED_ACTION_DB['definition'] = UPDATED_ACTION_DEFINITION
UPDATED_ACTION = copy.deepcopy(ACTION)
UPDATED_ACTION['definition'] = UPDATED_ACTION_DEFINITION

MOCK_ACTION = mock.MagicMock(return_value=ACTION_DB)
MOCK_SYSTEM_ACTION = mock.MagicMock(return_value=SYSTEM_ACTION_DB)
MOCK_ACTIONS = mock.MagicMock(return_value=[ACTION_DB])
MOCK_UPDATED_ACTION = mock.MagicMock(return_value=UPDATED_ACTION_DB)
MOCK_DELETE = mock.MagicMock(return_value=None)
MOCK_EMPTY = mock.MagicMock(return_value=[])
MOCK_NOT_FOUND = mock.MagicMock(side_effect=exc.DBEntityNotFoundError())
MOCK_DUPLICATE = mock.MagicMock(side_effect=exc.DBDuplicateEntryError())


class TestActionsController(base.APITest):
    @mock.patch.object(
        db_api, "get_action_definition", MOCK_ACTION)
    def test_get(self):
        resp = self.app.get('/v2/actions/my_action')

        self.assertEqual(200, resp.status_int)
        self.assertDictEqual(ACTION, resp.json)

    @mock.patch.object(
        db_api, "get_action_definition", MOCK_NOT_FOUND)
    def test_get_not_found(self):
        resp = self.app.get('/v2/actions/my_action', expect_errors=True)

        self.assertEqual(404, resp.status_int)

    @mock.patch.object(db_api, "update_action_definition", MOCK_UPDATED_ACTION)
    @mock.patch.object(
        db_api, "get_action_definition", MOCK_ACTION)
    def test_get_by_id(self):
        url = '/v2/actions/{0}'.format(ACTION['id'])
        resp = self.app.get(url)

        self.assertEqual(200, resp.status_int)
        self.assertEqual(ACTION['id'], resp.json['id'])

    @mock.patch.object(
        db_api, "get_action_definition", MOCK_NOT_FOUND)
    def test_get_by_id_not_found(self):
        url = '/v2/actions/1234'
        resp = self.app.get(url, expect_errors=True)

        self.assertEqual(404, resp.status_int)

    @mock.patch.object(
        db_api, "get_action_definition", MOCK_ACTION)
    @mock.patch.object(
        db_api, "update_action_definition", MOCK_UPDATED_ACTION
    )
    def test_put(self):
        resp = self.app.put(
            '/v2/actions',
            UPDATED_ACTION_DEFINITION,
            headers={'Content-Type': 'text/plain'}
        )

        self.assertEqual(200, resp.status_int)

        self.assertEqual({"actions": [UPDATED_ACTION]}, resp.json)

    @mock.patch.object(db_api, "load_action_definition", MOCK_ACTION)
    @mock.patch.object(db_api, "update_action_definition")
    def test_put_public(self, mock_mtd):
        mock_mtd.return_value = UPDATED_ACTION_DB

        resp = self.app.put(
            '/v2/actions?scope=public',
            UPDATED_ACTION_DEFINITION,
            headers={'Content-Type': 'text/plain'}
        )

        self.assertEqual(200, resp.status_int)

        self.assertEqual({"actions": [UPDATED_ACTION]}, resp.json)

        self.assertEqual("public", mock_mtd.call_args[0][1]['scope'])

    @mock.patch.object(db_api, "update_action_definition", MOCK_NOT_FOUND)
    def test_put_not_found(self):
        resp = self.app.put(
            '/v2/actions',
            UPDATED_ACTION_DEFINITION,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(404, resp.status_int)

    @mock.patch.object(
        db_api, "get_action_definition", MOCK_SYSTEM_ACTION)
    def test_put_system(self):
        resp = self.app.put(
            '/v2/actions',
            SYSTEM_ACTION_DEFINITION,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)
        self.assertIn(
            'Attempt to modify a system action: std.echo',
            resp.body.decode()
        )

    @mock.patch.object(db_api, "create_action_definition")
    def test_post(self, mock_mtd):
        mock_mtd.return_value = ACTION_DB

        resp = self.app.post(
            '/v2/actions',
            ACTION_DEFINITION,
            headers={'Content-Type': 'text/plain'}
        )

        self.assertEqual(201, resp.status_int)
        self.assertEqual({"actions": [ACTION]}, resp.json)

        self.assertEqual(1, mock_mtd.call_count)

        values = mock_mtd.call_args[0][0]

        self.assertEqual('My super cool action.', values['description'])

        spec = values['spec']

        self.assertIsNotNone(spec)
        self.assertEqual(ACTION_DB.name, spec['name'])

    @mock.patch.object(db_api, "create_action_definition")
    def test_post_public(self, mock_mtd):
        mock_mtd.return_value = ACTION_DB

        resp = self.app.post(
            '/v2/actions?scope=public',
            ACTION_DEFINITION,
            headers={'Content-Type': 'text/plain'}
        )

        self.assertEqual(201, resp.status_int)
        self.assertEqual({"actions": [ACTION]}, resp.json)

        self.assertEqual("public", mock_mtd.call_args[0][0]['scope'])

    @mock.patch.object(db_api, "create_action_definition")
    def test_post_wrong_scope(self, mock_mtd):
        mock_mtd.return_value = ACTION_DB

        resp = self.app.post(
            '/v2/actions?scope=unique',
            ACTION_DEFINITION,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)
        self.assertIn("Scope must be one of the following", resp.body.decode())

    @mock.patch.object(db_api, "create_action_definition", MOCK_DUPLICATE)
    def test_post_dup(self):
        resp = self.app.post(
            '/v2/actions',
            ACTION_DEFINITION,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(409, resp.status_int)

    @mock.patch.object(
        db_api, "get_action_definition", MOCK_ACTION)
    @mock.patch.object(db_api, "delete_action_definition", MOCK_DELETE)
    def test_delete(self):
        resp = self.app.delete('/v2/actions/my_action')

        self.assertEqual(204, resp.status_int)

    @mock.patch.object(db_api, "delete_action_definition", MOCK_NOT_FOUND)
    def test_delete_not_found(self):
        resp = self.app.delete('/v2/actions/my_action', expect_errors=True)

        self.assertEqual(404, resp.status_int)

    @mock.patch.object(
        db_api, "get_action_definition", MOCK_SYSTEM_ACTION)
    def test_delete_system(self):
        resp = self.app.delete('/v2/actions/std.echo', expect_errors=True)

        self.assertEqual(400, resp.status_int)
        self.assertIn('Attempt to delete a system action: std.echo',
                      resp.json['faultstring'])

    @mock.patch.object(
        db_api, "get_action_definitions", MOCK_ACTIONS)
    def test_get_all(self):
        resp = self.app.get('/v2/actions')

        self.assertEqual(200, resp.status_int)

        self.assertEqual(1, len(resp.json['actions']))
        self.assertDictEqual(ACTION, resp.json['actions'][0])

    @mock.patch.object(
        db_api, "get_action_definitions", MOCK_EMPTY)
    def test_get_all_empty(self):
        resp = self.app.get('/v2/actions')

        self.assertEqual(200, resp.status_int)

        self.assertEqual(0, len(resp.json['actions']))

    @mock.patch.object(
        db_api, "get_action_definitions", MOCK_ACTIONS)
    def test_get_all_pagination(self):
        resp = self.app.get(
            '/v2/actions?limit=1&sort_keys=id,name')

        self.assertEqual(200, resp.status_int)
        self.assertIn('next', resp.json)
        self.assertEqual(1, len(resp.json['actions']))
        self.assertDictEqual(ACTION, resp.json['actions'][0])

        param_dict = utils.get_dict_from_string(
            resp.json['next'].split('?')[1],
            delimiter='&'
        )

        expected_dict = {
            'marker': '123e4567-e89b-12d3-a456-426655440000',
            'limit': 1,
            'sort_keys': 'id,name',
            'sort_dirs': 'asc,asc'
        }

        self.assertTrue(
            set(expected_dict.items()).issubset(set(param_dict.items()))
        )

    def test_get_all_pagination_limit_negative(self):
        resp = self.app.get(
            '/v2/actions?limit=-1&sort_keys=id,name&sort_dirs=asc,asc',
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)

        self.assertIn("Limit must be positive", resp.body.decode())

    def test_get_all_pagination_limit_not_integer(self):
        resp = self.app.get(
            '/v2/actions?limit=1.1&sort_keys=id,name&sort_dirs=asc,asc',
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)

        self.assertIn("unable to convert to int", resp.body.decode())

    def test_get_all_pagination_invalid_sort_dirs_length(self):
        resp = self.app.get(
            '/v2/actions?limit=1&sort_keys=id,name&sort_dirs=asc,asc,asc',
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)

        self.assertIn(
            "Length of sort_keys must be equal or greater than sort_dirs",
            resp.body.decode()
        )

    def test_get_all_pagination_unknown_direction(self):
        resp = self.app.get(
            '/v2/actions?limit=1&sort_keys=id&sort_dirs=nonexist',
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)

        self.assertIn("Unknown sort direction", resp.body.decode())

    def test_validate(self):
        resp = self.app.post(
            '/v2/actions/validate',
            ACTION_DEFINITION,
            headers={'Content-Type': 'text/plain'}
        )

        self.assertEqual(200, resp.status_int)
        self.assertTrue(resp.json['valid'])

    def test_validate_invalid_model_exception(self):
        resp = self.app.post(
            '/v2/actions/validate',
            ACTION_DEFINITION_INVALID_NO_BASE,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(200, resp.status_int)
        self.assertFalse(resp.json['valid'])
        self.assertIn("Invalid DSL", resp.json['error'])

    def test_validate_dsl_parse_exception(self):
        resp = self.app.post(
            '/v2/actions/validate',
            ACTION_DSL_PARSE_EXCEPTION,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(200, resp.status_int)
        self.assertFalse(resp.json['valid'])
        self.assertIn("Definition could not be parsed", resp.json['error'])

    def test_validate_yaql_parse_exception(self):
        resp = self.app.post(
            '/v2/actions/validate',
            ACTION_DEFINITION_INVALID_YAQL,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(200, resp.status_int)
        self.assertFalse(resp.json['valid'])
        self.assertIn("unexpected end of statement",
                      resp.json['error'])

    def test_validate_empty(self):
        resp = self.app.post(
            '/v2/actions/validate',
            '',
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(200, resp.status_int)
        self.assertFalse(resp.json['valid'])
        self.assertIn("Invalid DSL", resp.json['error'])



# Copyright 2016 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Add named lock table

Revision ID: 017
Revises: 016
Create Date: 2016-08-17 13:06:26.616451

"""

# revision identifiers, used by Alembic.
revision = '017'
down_revision = '016'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'named_locks',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=250), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name')
    )



# Copyright 2014 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import abc

from oslo_serialization import jsonutils


class Serializer(object):
    @staticmethod
    @abc.abstractmethod
    def serialize(entity):
        pass

    @staticmethod
    @abc.abstractmethod
    def deserialize(entity):
        pass


class KombuSerializer(Serializer):
    @staticmethod
    def deserialize(entity):
        return jsonutils.loads(entity)

    @staticmethod
    def serialize(entity):
        return jsonutils.dumps(
            jsonutils.to_primitive(entity, convert_instances=True)
        )



# Copyright 2015 - Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import oslo_messaging as messaging

from mistral import context as auth_ctx
from mistral.engine.rpc_backend import base as rpc_base
from mistral.engine.rpc_backend import rpc


class OsloRPCClient(rpc_base.RPCClient):
    def __init__(self, conf):
        super(OsloRPCClient, self).__init__(conf)
        self.topic = conf.topic

        serializer = auth_ctx.RpcContextSerializer(
            auth_ctx.JsonPayloadSerializer())

        self._client = messaging.RPCClient(
            rpc.get_transport(),
            messaging.Target(topic=self.topic),
            serializer=serializer
        )

    def sync_call(self, ctx, method, target=None, **kwargs):
        return self._client.prepare(topic=self.topic, server=target).call(
            ctx,
            method,
            **kwargs
        )

    def async_call(self, ctx, method, target=None, **kwargs):
        return self._client.prepare(topic=self.topic, server=target).cast(
            ctx,
            method,
            **kwargs
        )



# Copyright 2016 - Catalyst IT Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy

import mock
from oslo_config import cfg
from oslo_utils import uuidutils

from mistral.db.v2 import api as db_api
from mistral.services import security
from mistral.tests.unit.api import base

GET_PROJECT_PATH = 'mistral.services.security.get_project_id'

WF_DEFINITION = {
    'name': 'test_wf',
    'definition': 'empty',
    'spec': {},
    'tags': ['mc'],
    'scope': 'private',
    'project_id': security.get_project_id(),
    'trust_id': '1234'
}

WORKFLOW_MEMBER_PENDING = {
    'member_id': '11-22-33',
    'project_id': '<default-project>',
    'resource_type': 'workflow',
    'status': 'pending'
}

WORKFLOW_MEMBER_ACCEPTED = {}

MEMBER_URL = None


class TestMembersController(base.APITest):
    def setUp(self):
        super(TestMembersController, self).setUp()

        wf = db_api.create_workflow_definition(WF_DEFINITION)

        global MEMBER_URL, WORKFLOW_MEMBER_ACCEPTED

        MEMBER_URL = '/v2/workflows/%s/members' % wf.id

        WORKFLOW_MEMBER_PENDING['resource_id'] = wf.id

        WORKFLOW_MEMBER_ACCEPTED = copy.deepcopy(WORKFLOW_MEMBER_PENDING)
        WORKFLOW_MEMBER_ACCEPTED['status'] = 'accepted'

        cfg.CONF.set_default('auth_enable', True, group='pecan')

    def test_membership_api_without_auth(self):
        cfg.CONF.set_default('auth_enable', False, group='pecan')

        resp = self.app.get(MEMBER_URL, expect_errors=True)

        self.assertEqual(400, resp.status_int)
        self.assertIn(
            "Resource sharing feature can only be supported with "
            "authentication enabled",
            resp.body.decode()
        )

    @mock.patch('mistral.context.AuthHook.before')
    def test_create_resource_member(self, auth_mock):
        # Workflow owner shares workflow to another tenant.
        resp = self.app.post_json(MEMBER_URL, {'member_id': '11-22-33'})

        self.assertEqual(201, resp.status_int)
        self._assert_dict_contains_subset(WORKFLOW_MEMBER_PENDING, resp.json)

    @mock.patch('mistral.context.AuthHook.before')
    def test_create_membership_nonexistent_wf(self, auth_mock):
        nonexistent_wf_id = uuidutils.generate_uuid()

        resp = self.app.post_json(
            '/v2/workflows/%s/members' % nonexistent_wf_id,
            {'member_id': '11-22-33'},
            expect_errors=True
        )

        self.assertEqual(404, resp.status_int)

    @mock.patch('mistral.context.AuthHook.before')
    def test_create_duplicate_membership(self, auth_mock):
        resp = self.app.post_json(MEMBER_URL, {'member_id': '11-22-33'})

        self.assertEqual(201, resp.status_int)

        resp = self.app.post_json(
            MEMBER_URL,
            {'member_id': '11-22-33'},
            expect_errors=True
        )

        self.assertEqual(409, resp.status_int)
        self.assertIn("Duplicate entry for ResourceMember", resp.body.decode())

    @mock.patch('mistral.context.AuthHook.before')
    def test_create_membership_public_wf(self, auth_mock):
        WF_DEFINITION_PUBLIC = copy.deepcopy(WF_DEFINITION)
        WF_DEFINITION_PUBLIC['name'] = 'test_wf1'
        WF_DEFINITION_PUBLIC['scope'] = 'public'

        wf_public = db_api.create_workflow_definition(WF_DEFINITION_PUBLIC)

        resp = self.app.post_json(
            '/v2/workflows/%s/members' % wf_public.id,
            {'member_id': '11-22-33'},
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)
        self.assertIn(
            "Only private resource could be shared",
            resp.body.decode()
        )

    @mock.patch('mistral.context.AuthHook.before')
    def test_create_membership_untransferable(self, auth_mock):
        resp = self.app.post_json(MEMBER_URL, {'member_id': '11-22-33'})

        self.assertEqual(201, resp.status_int)

        # Using mock to switch to another tenant.
        get_mock = mock.MagicMock(return_value='11-22-33')

        with mock.patch(GET_PROJECT_PATH, get_mock):
            resp = self.app.post_json(
                MEMBER_URL,
                {'member_id': 'other_tenant'},
                expect_errors=True
            )

        self.assertEqual(404, resp.status_int)

    @mock.patch('mistral.context.AuthHook.before')
    def test_get_other_memberships(self, auth_mock):
        resp = self.app.post_json(MEMBER_URL, {'member_id': '11-22-33'})

        self.assertEqual(201, resp.status_int)

        # Using mock to switch to another tenant.
        get_mock = mock.MagicMock(return_value='other_tenant')

        with mock.patch(GET_PROJECT_PATH, get_mock):
            resp = self.app.get(MEMBER_URL)

        self.assertEqual(200, resp.status_int)
        self.assertEqual(0, len(resp.json['members']))

    @mock.patch('mistral.context.AuthHook.before')
    def test_get_memberships_nonexistent_wf(self, auth_mock):
        nonexistent_wf_id = uuidutils.generate_uuid()

        resp = self.app.get(
            '/v2/workflows/%s/members' % nonexistent_wf_id,
        )

        self.assertEqual(200, resp.status_int)
        self.assertEqual(0, len(resp.json['members']))

    @mock.patch('mistral.context.AuthHook.before')
    def test_get_resource_memberips(self, auth_mock):
        # Workflow owner shares workflow to another tenant.
        resp = self.app.post_json(MEMBER_URL, {'member_id': '11-22-33'})

        self.assertEqual(201, resp.status_int)
        self._assert_dict_contains_subset(WORKFLOW_MEMBER_PENDING, resp.json)

        # Workflow owner queries the workflow members.
        resp = self.app.get(MEMBER_URL)

        self.assertEqual(200, resp.status_int)
        self.assertEqual(1, len(resp.json['members']))
        self._assert_dict_contains_subset(
            WORKFLOW_MEMBER_PENDING,
            resp.json['members'][0]
        )

        # Workflow owner queries the exact workflow member.
        resp = self.app.get('%s/11-22-33' % MEMBER_URL)

        self.assertEqual(200, resp.status_int)
        self._assert_dict_contains_subset(
            WORKFLOW_MEMBER_PENDING,
            resp.json
        )

    @mock.patch('mistral.context.AuthHook.before')
    def test_get_other_membership(self, auth_mock):
        resp = self.app.post_json(MEMBER_URL, {'member_id': '11-22-33'})

        self.assertEqual(201, resp.status_int)

        # Using mock to switch to another tenant.
        get_mock = mock.MagicMock(return_value='other_tenant')

        with mock.patch(GET_PROJECT_PATH, get_mock):
            resp = self.app.get(
                '%s/11-22-33' % MEMBER_URL,
                expect_errors=True
            )

        self.assertEqual(404, resp.status_int)

    @mock.patch('mistral.context.AuthHook.before')
    def test_update_membership(self, auth_mock):
        # Workflow owner shares workflow to another tenant.
        resp = self.app.post_json(MEMBER_URL, {'member_id': '11-22-33'})

        self.assertEqual(201, resp.status_int)

        # Using mock to switch to another tenant.
        get_mock = mock.MagicMock(return_value='11-22-33')

        # Tenant accepts the workflow shared to him.
        with mock.patch(GET_PROJECT_PATH, get_mock):
            resp = self.app.put_json(
                '%s/11-22-33' % MEMBER_URL,
                {'status': 'accepted'}
            )

        self.assertEqual(200, resp.status_int)
        self._assert_dict_contains_subset(
            WORKFLOW_MEMBER_ACCEPTED,
            resp.json
        )

        # Tenant queries exact member of workflow shared to him.
        # (status=accepted).
        with mock.patch(GET_PROJECT_PATH, get_mock):
            resp = self.app.get('%s/11-22-33' % MEMBER_URL)

        self.assertEqual(200, resp.status_int)
        self._assert_dict_contains_subset(
            WORKFLOW_MEMBER_ACCEPTED,
            resp.json
        )

        # Workflow owner queries the exact workflow member.
        # (status=accepted).
        resp = self.app.get('%s/11-22-33' % MEMBER_URL)

        self.assertEqual(200, resp.status_int)
        self._assert_dict_contains_subset(
            WORKFLOW_MEMBER_ACCEPTED,
            resp.json
        )

    @mock.patch('mistral.context.AuthHook.before')
    def test_update_membership_invalid_status(self, auth_mock):
        resp = self.app.post_json(MEMBER_URL, {'member_id': '11-22-33'})

        self.assertEqual(201, resp.status_int)

        # Using mock to switch to another tenant.
        get_mock = mock.MagicMock(return_value='11-22-33')

        with mock.patch(GET_PROJECT_PATH, get_mock):
            resp = self.app.put_json(
                '%s/11-22-33' % MEMBER_URL,
                {'status': 'invalid'},
                expect_errors=True
            )

        self.assertEqual(400, resp.status_int)
        self.assertIn(
            "Invalid input",
            resp.body.decode()
        )

    @mock.patch('mistral.context.AuthHook.before')
    def test_update_membership_not_shared_user(self, auth_mock):
        resp = self.app.post_json(MEMBER_URL, {'member_id': '11-22-33'})

        self.assertEqual(201, resp.status_int)

        resp = self.app.put_json(
            '%s/11-22-33' % MEMBER_URL,
            {'status': 'accepted'},
            expect_errors=True
        )

        self.assertEqual(404, resp.status_int)

    @mock.patch('mistral.context.AuthHook.before')
    def test_delete_membership(self, auth_mock):
        # Workflow owner shares workflow to another tenant.
        resp = self.app.post_json(MEMBER_URL, {'member_id': '11-22-33'})

        self.assertEqual(201, resp.status_int)

        # Workflow owner deletes the exact workflow member.
        resp = self.app.delete('%s/11-22-33' % MEMBER_URL)

        self.assertEqual(204, resp.status_int)

        # Workflow owner queries the workflow members.
        resp = self.app.get(MEMBER_URL)

        self.assertEqual(200, resp.status_int)
        self.assertEqual(0, len(resp.json['members']))

        # Using mock to switch to another tenant.
        get_mock = mock.MagicMock(return_value='11-22-33')

        # Tenant queries members of workflow shared to him, after deletion.
        with mock.patch(GET_PROJECT_PATH, get_mock):
            resp = self.app.get(MEMBER_URL)

        self.assertEqual(200, resp.status_int)
        self.assertEqual(0, len(resp.json['members']))

    @mock.patch('mistral.context.AuthHook.before')
    def test_delete_membership_not_owner(self, auth_mock):
        resp = self.app.post_json(MEMBER_URL, {'member_id': '11-22-33'})

        self.assertEqual(201, resp.status_int)

        # Using mock to switch to another tenant.
        get_mock = mock.MagicMock(return_value='11-22-33')

        with mock.patch(GET_PROJECT_PATH, get_mock):
            resp = self.app.delete(
                '%s/11-22-33' % MEMBER_URL,
                expect_errors=True
            )

        self.assertEqual(404, resp.status_int)



# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import datetime
import mock
from oslo_config import cfg
import pecan
import pecan.testing
import requests_mock

from mistral.db.v2 import api as db_api
from mistral.db.v2.sqlalchemy import models
from mistral.services import periodic
from mistral.tests.unit import base
from mistral.tests.unit.mstrlfixtures import policy_fixtures


WF_DEFINITION = """
---
version: '2.0'

flow:
  type: direct
  input:
    - param1

  tasks:
    task1:
      action: std.echo output="Hi"
"""

WF_DB = models.WorkflowDefinition(
    id='123e4567-e89b-12d3-a456-426655440000',
    name='flow',
    definition=WF_DEFINITION,
    created_at=datetime.datetime(1970, 1, 1),
    updated_at=datetime.datetime(1970, 1, 1),
    spec={'input': ['param1']}
)

WF = {
    'id': '123e4567-e89b-12d3-a456-426655440000',
    'name': 'flow',
    'definition': WF_DEFINITION,
    'created_at': '1970-01-01 00:00:00',
    'updated_at': '1970-01-01 00:00:00',
    'input': 'param1'
}


MOCK_WF = mock.MagicMock(return_value=WF_DB)

# Set up config options.
AUTH_URL = 'https://my.keycloak.com:8443/auth'
REALM_NAME = 'my_realm'

USER_INFO_ENDPOINT = (
    "%s/realms/%s/protocol/openid-connect/userinfo" % (AUTH_URL, REALM_NAME)
)

USER_CLAIMS = {
    "sub": "248289761001",
    "name": "Jane Doe",
    "given_name": "Jane",
    "family_name": "Doe",
    "preferred_username": "j.doe",
    "email": "janedoe@example.com",
    "picture": "http://example.com/janedoe/me.jpg"
}


class TestKeyCloakOIDCAuth(base.DbTestCase):
    def setUp(self):
        super(TestKeyCloakOIDCAuth, self).setUp()

        cfg.CONF.set_default('auth_enable', True, group='pecan')
        cfg.CONF.set_default('auth_type', 'keycloak-oidc')
        cfg.CONF.set_default('auth_url', AUTH_URL, group='keycloak_oidc')

        pecan_opts = cfg.CONF.pecan

        self.app = pecan.testing.load_test_app({
            'app': {
                'root': pecan_opts.root,
                'modules': pecan_opts.modules,
                'debug': pecan_opts.debug,
                'auth_enable': True,
                'disable_cron_trigger_thread': True
            }
        })

        self.addCleanup(pecan.set_config, {}, overwrite=True)
        self.addCleanup(
            cfg.CONF.set_default,
            'auth_enable',
            False,
            group='pecan'
        )
        self.addCleanup(cfg.CONF.set_default, 'auth_type', 'keystone')

        # Adding cron trigger thread clean up explicitly in case if
        # new tests will provide an alternative configuration for pecan
        # application.
        self.addCleanup(periodic.stop_all_periodic_tasks)

        # Make sure the api get the correct context.
        self.patch_ctx = mock.patch(
            'mistral.context.context_from_headers_and_env'
        )
        self.mock_ctx = self.patch_ctx.start()
        self.mock_ctx.return_value = self.ctx
        self.addCleanup(self.patch_ctx.stop)

        self.policy = self.useFixture(policy_fixtures.PolicyFixture())

    @requests_mock.Mocker()
    @mock.patch.object(db_api, 'get_workflow_definition', MOCK_WF)
    def test_get_workflow_success_auth(self, req_mock):
        # Imitate successful response from KeyCloak with user claims.
        req_mock.get(USER_INFO_ENDPOINT, json=USER_CLAIMS)

        headers = {
            'X-Auth-Token': 'cvbcvbasrtqlwkjasdfasdf',
            'X-Project-Id': REALM_NAME
        }

        resp = self.app.get('/v2/workflows/123', headers=headers)

        self.assertEqual(200, resp.status_code)
        self.assertDictEqual(WF, resp.json)

    @requests_mock.Mocker()
    @mock.patch.object(db_api, 'get_workflow_definition', MOCK_WF)
    def test_get_workflow_failed_auth(self, req_mock):
        # Imitate failure response from KeyCloak.
        req_mock.get(
            USER_INFO_ENDPOINT,
            status_code=401,
            reason='Access token is invalid'
        )

        headers = {
            'X-Auth-Token': 'cvbcvbasrtqlwkjasdfasdf',
            'X-Project-Id': REALM_NAME
        }

        resp = self.app.get(
            '/v2/workflows/123',
            headers=headers,
            expect_errors=True
        )

        self.assertEqual(401, resp.status_code)
        self.assertEqual('401 Unauthorized', resp.status)
        self.assertIn('Failed to validate access token', resp.text)
        self.assertIn('Access token is invalid', resp.text)



# Copyright 2014 - Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_config import cfg
import testtools

from mistral.db.v2 import api as db_api
from mistral.services import workflows as wf_service
from mistral.tests.unit.engine import base
from mistral.workflow import states


# Use the set_default method to set value otherwise in certain test cases
# the change in value is not permanent.
cfg.CONF.set_default('auth_enable', False, group='pecan')


class JoinEngineTest(base.EngineTestCase):
    def test_full_join_simple(self):
        wf_text = """---
        version: '2.0'

        wf:
          type: direct

          tasks:
            join_task:
              join: all

            task1:
              on-success: join_task

            task2:
              on-success: join_task
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf', {})

        self.await_workflow_success(wf_ex.id)

    def test_full_join_without_errors(self):
        wf_text = """---
        version: '2.0'

        wf:
          type: direct

          output:
            result: <% $.result3 %>

          tasks:
            task1:
              action: std.echo output=1
              publish:
                result1: <% task(task1).result %>
              on-complete:
                - task3

            task2:
              action: std.echo output=2
              publish:
                result2: <% task(task2).result %>
              on-complete:
                - task3

            task3:
              join: all
              action: std.echo output="<% $.result1 %>,<% $.result2 %>"
              publish:
                result3: <% task(task3).result %>
        """

        wf_service.create_workflows(wf_text)

        # Start workflow.
        wf_ex = self.engine.start_workflow('wf', {})

        self.await_workflow_success(wf_ex.id)

        # Note: We need to reread execution to access related tasks.
        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            self.assertDictEqual({'result': '1,2'}, wf_ex.output)

            tasks = wf_ex.task_executions

        task1 = self._assert_single_item(tasks, name='task1')
        task2 = self._assert_single_item(tasks, name='task2')
        task3 = self._assert_single_item(tasks, name='task3')

        self.assertEqual(states.SUCCESS, task1.state)
        self.assertEqual(states.SUCCESS, task2.state)
        self.assertEqual(states.SUCCESS, task3.state)

    def test_full_join_with_errors(self):
        wf_text = """---
        version: '2.0'

        wf:
          type: direct

          output:
            result: <% $.result3 %>

          tasks:
            task1:
              action: std.echo output=1
              publish:
                result1: <% task(task1).result %>
              on-complete:
                - task3

            task2:
              action: std.fail
              on-error:
                - task3

            task3:
              join: all
              action: std.echo output="<% $.result1 %>-<% $.result1 %>"
              publish:
                result3: <% task(task3).result %>
        """

        wf_service.create_workflows(wf_text)

        # Start workflow.
        wf_ex = self.engine.start_workflow('wf', {})

        self.await_workflow_success(wf_ex.id)

        # Note: We need to reread execution to access related tasks.
        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            self.assertDictEqual({'result': '1-1'}, wf_ex.output)

            tasks = wf_ex.task_executions

        task1 = self._assert_single_item(tasks, name='task1')
        task2 = self._assert_single_item(tasks, name='task2')
        task3 = self._assert_single_item(tasks, name='task3')

        self.assertEqual(states.SUCCESS, task1.state)
        self.assertEqual(states.ERROR, task2.state)
        self.assertEqual(states.SUCCESS, task3.state)

    def test_full_join_with_conditions(self):
        wf_text = """---
        version: '2.0'

        wf:
          type: direct

          output:
            result: <% $.result4 %>

          tasks:
            task1:
              action: std.echo output=1
              publish:
                result1: <% task(task1).result %>
              on-complete:
                - task3

            task2:
              action: std.echo output=2
              publish:
                result2: <% task(task2).result %>
              on-complete:
                - task3: <% $.result2 = 11111 %>
                - task4: <% $.result2 = 2 %>

            task3:
              join: all
              action: std.echo output="<% $.result1 %>-<% $.result1 %>"
              publish:
                result3: <% task(task3).result %>

            task4:
              action: std.echo output=4
              publish:
                result4: <% task(task4).result %>
        """

        wf_service.create_workflows(wf_text)

        # Start workflow.
        wf_ex = self.engine.start_workflow('wf', {})

        def _num_of_tasks():
            return len(
                db_api.get_task_executions(workflow_execution_id=wf_ex.id)
            )

        self._await(lambda: _num_of_tasks() == 4)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            tasks = wf_ex.task_executions

        task1 = self._assert_single_item(tasks, name='task1')
        task2 = self._assert_single_item(tasks, name='task2')
        task3 = self._assert_single_item(tasks, name='task3')
        task4 = self._assert_single_item(tasks, name='task4')

        # NOTE(xylan): We ensure task4 is successful here because of the
        # uncertainty of its running in parallel with task3.
        self.await_task_success(task4.id)

        self.assertEqual(states.RUNNING, wf_ex.state)
        self.assertEqual(states.SUCCESS, task1.state)
        self.assertEqual(states.SUCCESS, task2.state)

        # NOTE(rakhmerov): Task 3 must fail because task2->task3 transition
        # will never trigger due to its condition.
        self.await_task_error(task3.id)

        self.await_workflow_error(wf_ex.id)

    def test_partial_join(self):
        wf_text = """---
        version: '2.0'

        wf:
          type: direct

          output:
            result: <% $.result4 %>

          tasks:
            task1:
              action: std.echo output=1
              publish:
                result1: <% task(task1).result %>
              on-complete:
                - task4

            task2:
              action: std.echo output=2
              publish:
                result2: <% task(task2).result %>
              on-complete:
                - task4

            task3:
              action: std.fail
              description: |
                Always fails and 'on-success' never gets triggered.
                However, 'task4' will run since its join cardinality
                is 2 which means 'task1' and 'task2' completion is
                enough to trigger it.
              on-success:
                - task4
              on-error:
                - noop

            task4:
              join: 2
              action: std.echo output="<% $.result1 %>,<% $.result2 %>"
              publish:
                result4: <% task(task4).result %>
        """

        wf_service.create_workflows(wf_text)

        # Start workflow.
        wf_ex = self.engine.start_workflow('wf', {})

        self.await_workflow_success(wf_ex.id)

        # Note: We need to reread execution to access related tasks.
        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            self.assertDictEqual({'result': '1,2'}, wf_ex.output)

            tasks = wf_ex.task_executions

        self.assertEqual(4, len(tasks))

        task4 = self._assert_single_item(tasks, name='task4')
        task1 = self._assert_single_item(tasks, name='task1')
        task2 = self._assert_single_item(tasks, name='task2')
        task3 = self._assert_single_item(tasks, name='task3')

        self.assertEqual(states.SUCCESS, task1.state)
        self.assertEqual(states.SUCCESS, task2.state)
        self.assertEqual(states.SUCCESS, task4.state)

        # task3 may still be in RUNNING state and we need to make sure
        # it gets into ERROR state.
        self.await_task_error(task3.id)

        self.assertDictEqual({'result4': '1,2'}, task4.published)

    def test_partial_join_triggers_once(self):
        wf_text = """---
        version: '2.0'

        wf:
          type: direct

          output:
            result: <% $.result5 %>

          tasks:
            task1:
              action: std.noop
              publish:
                result1: 1
              on-complete:
                - task5

            task2:
              action: std.noop
              publish:
                result2: 2
              on-complete:
                - task5

            task3:
              action: std.noop
              publish:
                result3: 3
              on-complete:
                - task5

            task4:
              action: std.noop
              publish:
                result4: 4
              on-complete:
                - task5

            task5:
              join: 2
              action: std.echo
              input:
                output: |
                  <% result1 in $.keys() %>,<% result2 in $.keys() %>,
                  <% result3 in $.keys() %>,<% result4 in $.keys() %>
              publish:
                result5: <% task(task5).result %>
        """

        wf_service.create_workflows(wf_text)

        # Start workflow.
        wf_ex = self.engine.start_workflow('wf', {})

        self.await_workflow_success(wf_ex.id)

        # Note: We need to reread execution to access related tasks.
        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            tasks = wf_ex.task_executions

        self.assertEqual(5, len(tasks))

        task5 = self._assert_single_item(tasks, name='task5')

        self.assertEqual(states.SUCCESS, task5.state)

        success_count = sum([1 for t in tasks if t.state == states.SUCCESS])

        # At least task4 and two others must be successfully completed.
        self.assertGreaterEqual(success_count, 3)

        result5 = task5.published['result5']

        self.assertIsNotNone(result5)

        # Depending on how many inbound tasks completed before 'join'
        # task5 started it can get different inbound context with.
        # But at least two inbound results should be accessible at task5
        # which logically corresponds to 'join' cardinality 2.
        self.assertGreaterEqual(result5.count('True'), 2)

    def test_discriminator(self):
        wf_text = """---
        version: '2.0'

        wf:
          type: direct

          output:
            result: <% $.result4 %>

          tasks:
            task1:
              action: std.noop
              publish:
                result1: 1
              on-complete:
                - task4

            task2:
              action: std.noop
              publish:
                result2: 2
              on-complete:
                - task4

            task3:
              action: std.noop
              publish:
                result3: 3
              on-complete:
                - task4

            task4:
              join: one
              action: std.echo
              input:
                output: |
                  <% result1 in $.keys() %>,<% result2 in $.keys() %>,
                  <% result3 in $.keys() %>
              publish:
                result4: <% task(task4).result %>
        """

        wf_service.create_workflows(wf_text)

        # Start workflow.
        wf_ex = self.engine.start_workflow('wf', {})

        self.await_workflow_success(wf_ex.id)

        # Note: We need to reread execution to access related tasks.
        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            tasks = wf_ex.task_executions

        self.assertEqual(4, len(tasks))

        task4 = self._assert_single_item(tasks, name='task4')

        self.assertEqual(states.SUCCESS, task4.state)

        success_count = sum([1 for t in tasks if t.state == states.SUCCESS])

        # At least task4 and one of others must be successfully completed.
        self.assertGreaterEqual(success_count, 2)

        result4 = task4.published['result4']

        self.assertIsNotNone(result4)
        self.assertLess(result4.count('False'), 3)
        self.assertGreaterEqual(result4.count('True'), 1)

    def test_full_join_parallel_published_vars(self):
        wfs_tasks_join_complex = """---
        version: '2.0'

        main:
          type: direct

          output:
            var1: <% $.var1 %>
            var2: <% $.var2 %>
            is_done: <% $.is_done %>

          tasks:
            init:
              publish:
                var1: false
                var2: false
                is_done: false
              on-success:
                - branch1
                - branch2

            branch1:
              workflow: work
              publish:
                var1: true
              on-success:
                - done

            branch2:
              publish:
                var2: true
              on-success:
                - done

            done:
              join: all
              publish:
                is_done: true

        work:
          type: direct

          tasks:
            do:
              action: std.echo output="Doing..."
              on-success:
                - exit

            exit:
              action: std.echo output="Exiting..."
        """
        wf_service.create_workflows(wfs_tasks_join_complex)

        # Start workflow.
        wf_ex = self.engine.start_workflow('main', {})

        self.await_workflow_success(wf_ex.id)

        # Note: We need to reread execution to access related tasks.
        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            self.assertDictEqual(
                {
                    'var1': True,
                    'is_done': True,
                    'var2': True
                },
                wf_ex.output
            )

    @testtools.skip('https://bugs.launchpad.net/mistral/+bug/1424461')
    def test_full_join_parallel_published_vars_complex(self):
        wf_text = """---
        version: "2.0"

        main:
          type: direct
          output:
            var_a: <% $.var_a %>
            var_b: <% $.var_b %>
            var_c: <% $.var_c %>
            var_d: <% $.var_d %>
          tasks:
            init:
              publish:
                var_a: 0
                var_b: 0
                var_c: 0
              on-success:
                - branch1_0
                - branch2_0

            branch1_0:
              publish:
                var_c: 1
              on-success:
                - branch1_1

            branch2_0:
              publish:
                var_a: 1
              on-success:
                - done

            branch1_1:
              publish:
                var_b: 1
              on-success:
                - done

            done:
              join: all
              publish:
                var_d: 1
        """
        wf_service.create_workflows(wf_text)

        # Start workflow.
        wf_ex = self.engine.start_workflow('main', {})

        self.await_workflow_success(wf_ex.id)

        # Note: We need to reread execution to access related tasks.
        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            self.assertDictEqual(
                {
                    'var_a': 1,
                    'var_b': 1,
                    'var_c': 1,
                    'var_d': 1
                },
                wf_ex.output
            )

    def test_full_join_with_branch_errors(self):
        wf_text = """---
        version: '2.0'

        main:
          type: direct

          tasks:
            task10:
              action: std.noop
              on-success:
                - task21
                - task31

            task21:
              action: std.noop
              on-success:
                - task22

            task22:
              action: std.noop
              on-success:
                - task40

            task31:
              action: std.fail
              on-success:
                - task32

            task32:
              action: std.noop
              on-success:
                - task40

            task40:
              join: all
              action: std.noop
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('main', {})

        self.await_workflow_error(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            tasks = wf_ex.task_executions

        self.assertIsNotNone(wf_ex.state_info)

        task10 = self._assert_single_item(tasks, name='task10')
        task21 = self._assert_single_item(tasks, name='task21')
        task22 = self._assert_single_item(tasks, name='task22')
        task31 = self._assert_single_item(tasks, name='task31')
        task40 = self._assert_single_item(tasks, name='task40')

        self.assertEqual(states.SUCCESS, task10.state)
        self.assertEqual(states.SUCCESS, task21.state)
        self.assertEqual(states.SUCCESS, task22.state)
        self.assertEqual(states.ERROR, task31.state)
        self.assertNotIn('task32', [task.name for task in tasks])
        self.assertEqual(states.ERROR, task40.state)

    def test_diamond_join_all(self):
        wf_text = """---
        version: '2.0'

        test-join:
          tasks:
            a:
              on-success:
                - b
                - c
                - d

            b:
              on-success:
                - e

            c:
              on-success:
                - e

            d:
              on-success:
                - e

            e:
              join: all
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('test-join', {})

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            tasks = wf_ex.task_executions

        self._assert_multiple_items(tasks, 5, state=states.SUCCESS)

    def test_join_multiple_routes_with_one_source(self):
        wf_text = """---
        version: '2.0'

        wf:
          tasks:
            a:
              on-success:
                - b
                - c

            b:
              on-success:
                - c

            c:
              join: all
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf', {})

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            tasks = wf_ex.task_executions

            self._assert_multiple_items(tasks, 3, state=states.SUCCESS)

    def test_join_after_join(self):
        wf_text = """---
        version: '2.0'

        wf:
          tasks:
            a:
              on-success:
                - c

            b:
              on-success:
                - c

            c:
              join: all
              on-success:
                - f

            d:
              on-success:
                - f

            e:
              on-success:
                - f

            f:
              join: all
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf', {})

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)
            task_execs = wf_ex.task_executions

        self.assertEqual(6, len(task_execs))

        self._assert_multiple_items(task_execs, 6, state=states.SUCCESS)

    def test_join_route_delays(self):
        wf_text = """---
        version: '2.0'

        wf:
          tasks:
            a:
              wait-before: 4
              on-success: b
            b:
              on-success: join

            c:
              on-success: join

            join:
              join: all
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf', {})

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)
            task_execs = wf_ex.task_executions

        self.assertEqual(4, len(task_execs))

        self._assert_multiple_items(task_execs, 4, state=states.SUCCESS)

    def test_delete_join_completion_check_on_stop(self):
        wf_text = """---
        version: '2.0'

        wf:
          tasks:
            task1:
              action: std.noop
              on-success: join_task

            task2:
              description: Never ends
              action: std.async_noop
              on-success: join_task

            join_task:
              join: all
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf', {})

        tasks = db_api.get_task_executions(workflow_execution_id=wf_ex.id)

        self.assertTrue(len(tasks) >= 2)

        task1 = self._assert_single_item(tasks, name='task1')

        self.await_task_success(task1.id)

        # Once task1 is finished we know that join_task must be created.

        tasks = db_api.get_task_executions(workflow_execution_id=wf_ex.id)

        self._assert_single_item(
            tasks,
            name='join_task',
            state=states.WAITING
        )

        calls = db_api.get_delayed_calls()

        mtd_name = 'mistral.engine.task_handler._refresh_task_state'

        cnt = sum([1 for c in calls if c.target_method_name == mtd_name])

        # There can be 2 calls with different value of 'processing' flag.
        self.assertTrue(cnt == 1 or cnt == 2)

        # Stop the workflow.
        self.engine.stop_workflow(wf_ex.id, state=states.CANCELLED)

        self._await(
            lambda:
            len(db_api.get_delayed_calls(target_method_name=mtd_name)) == 0
        )

    def test_delete_join_completion_check_on_execution_delete(self):
        wf_text = """---
        version: '2.0'

        wf:
          tasks:
            task1:
              action: std.noop
              on-success: join_task

            task2:
              description: Never ends
              action: std.async_noop
              on-success: join_task

            join_task:
              join: all
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf', {})

        tasks = db_api.get_task_executions(workflow_execution_id=wf_ex.id)

        self.assertTrue(len(tasks) >= 2)

        task1 = self._assert_single_item(tasks, name='task1')

        self.await_task_success(task1.id)

        # Once task1 is finished we know that join_task must be created.

        tasks = db_api.get_task_executions(workflow_execution_id=wf_ex.id)

        self._assert_single_item(
            tasks,
            name='join_task',
            state=states.WAITING
        )

        calls = db_api.get_delayed_calls()

        mtd_name = 'mistral.engine.task_handler._refresh_task_state'

        cnt = sum([1 for c in calls if c.target_method_name == mtd_name])

        # There can be 2 calls with different value of 'processing' flag.
        self.assertTrue(cnt == 1 or cnt == 2)

        # Stop the workflow.
        db_api.delete_workflow_execution(wf_ex.id)

        self._await(
            lambda:
            len(db_api.get_delayed_calls(target_method_name=mtd_name)) == 0
        )

    def test_join_with_deep_dependencies_tree(self):
        wf_text = """---
        version: '2.0'

        wf:
          tasks:
            task_a_1:
              on-success:
                - task_with_join

            task_b_1:
              action: std.fail
              on-success:
                - task_b_2

            task_b_2:
              on-success:
                - task_b_3

            task_b_3:
              on-success:
                - task_with_join

            task_with_join:
              join: all
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf', {})

        self.await_workflow_error(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            task_execs = wf_ex.task_executions

        self.assertEqual(3, len(task_execs))
        self._assert_single_item(
            task_execs,
            name='task_a_1',
            state=states.SUCCESS
        )
        self._assert_single_item(
            task_execs,
            name='task_b_1',
            state=states.ERROR
        )
        self._assert_single_item(
            task_execs,
            name='task_with_join',
            state=states.ERROR
        )

    def test_no_workflow_error_after_inbound_error(self):
        wf_text = """---
        version: "2.0"

        wf:
          output:
            continue_flag: <% $.get(continue_flag) %>

          task-defaults:
            on-error:
              - change_continue_flag

          tasks:
            task_a:
              action: std.fail
              on-success:
                - task_c: <% $.get(continue_flag) = null %>
                - task_a_process

            task_a_process:
              action: std.noop

            task_b:
              on-success:
                - task_c: <% $.get(continue_flag) = null %>

            task_c:
              join: all

            change_continue_flag:
              publish:
                continue_flag: false
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf', {})

        self.await_workflow_success(wf_ex.id)



# Copyright 2015 - Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import sys

from mistral.engine.rpc_backend.kombu import kombu_client


# Example of using Kombu based RPC client.
def main():
    conf = {
        'user_id': 'guest',
        'password': 'secret',
        'exchange': 'my_exchange',
        'topic': 'my_topic',
        'server_id': 'host',
        'host': 'localhost',
        'port': 5672,
        'virtual_host': '/'
    }
    kombu_rpc = kombu_client.KombuRPCClient(conf)

    print(" [x] Requesting ...")

    ctx = type('context', (object,), {'to_dict': lambda self: {}})()

    response = kombu_rpc.sync_call(ctx, 'fib', n=44)

    print(" [.] Got %r" % (response,))


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2014 - Mirantis, Inc.
# Copyright 2015 - Huawei Technologies Co. Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import copy

from mistral.db.v2 import api as db_api
from mistral import exceptions as exc
from mistral import utils


# TODO(rakhmerov): This method is too abstract, validation rules may vary
#  depending on object type (action, wf), it's not clear what it can be
# applied to.
# TODO(rakhmerov): It must not do any manipulations with parameters
#  (input_dict)!
def validate_input(definition, input_dict, spec=None):
    input_param_names = copy.deepcopy(list((input_dict or {}).keys()))
    missing_param_names = []

    spec_input = (spec.get_input() if spec else
                  utils.get_dict_from_string(definition.input))

    for p_name, p_value in spec_input.items():
        if p_value is utils.NotDefined and p_name not in input_param_names:
            missing_param_names.append(str(p_name))

        if p_name in input_param_names:
            input_param_names.remove(p_name)

    if missing_param_names or input_param_names:
        msg = 'Invalid input [name=%s, class=%s'
        msg_props = [definition.name, spec.__class__.__name__]

        if missing_param_names:
            msg += ', missing=%s'
            msg_props.append(missing_param_names)

        if input_param_names:
            msg += ', unexpected=%s'
            msg_props.append(input_param_names)

        msg += ']'

        raise exc.InputException(
            msg % tuple(msg_props)
        )
    else:
        utils.merge_dicts(input_dict, spec_input, overwrite=False)


def resolve_workflow_definition(parent_wf_name, parent_wf_spec_name,
                                wf_spec_name):
    wf_def = None

    if parent_wf_name != parent_wf_spec_name:
        # If parent workflow belongs to a workbook then
        # check child workflow within the same workbook
        # (to be able to use short names within workbooks).
        # If it doesn't exist then use a name from spec
        # to find a workflow in DB.
        wb_name = parent_wf_name.rstrip(parent_wf_spec_name)[:-1]

        wf_full_name = "%s.%s" % (wb_name, wf_spec_name)

        wf_def = db_api.load_workflow_definition(wf_full_name)

    if not wf_def:
        wf_def = db_api.load_workflow_definition(wf_spec_name)

    if not wf_def:
        raise exc.WorkflowException(
            "Failed to find workflow [name=%s]" % wf_spec_name
        )

    return wf_def



# Copyright 2016 NEC Corporation. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
from tempest.lib import exceptions
from tempest import test

from mistral import utils
from mistral_tempest_tests.tests import base


class ActionTestsV2(base.TestCase):

    _service = 'workflowv2'

    def get_field_value(self, body, act_name, field):
        return [body['actions'][i][field]
                for i in range(len(body['actions']))
                if body['actions'][i]['name'] == act_name][0]

    def tearDown(self):
        for act in self.client.actions:
            self.client.delete_obj('actions', act)
        self.client.actions = []

        super(ActionTestsV2, self).tearDown()

    @test.attr(type='smoke')
    def test_get_list_actions(self):
        resp, body = self.client.get_list_obj('actions')

        self.assertEqual(200, resp.status)
        self.assertNotEqual([], body['actions'])
        self.assertNotIn('next', body)

    @test.attr(type='smoke')
    def test_get_list_actions_with_pagination(self):
        resp, body = self.client.get_list_obj(
            'actions?limit=1&sort_keys=name&sort_dirs=desc'
        )

        self.assertEqual(200, resp.status)
        self.assertEqual(1, len(body['actions']))
        self.assertIn('next', body)

        name_1 = body['actions'][0].get('name')
        next = body.get('next')

        param_dict = utils.get_dict_from_string(
            next.split('?')[1],
            delimiter='&'
        )

        # NOTE: 'id' gets included into sort keys automatically with 'desc'
        # sorting to avoid pagination looping.
        expected_sub_dict = {
            'limit': 1,
            'sort_keys': 'name,id',
            'sort_dirs': 'desc,asc'
        }

        self.assertDictContainsSubset(expected_sub_dict, param_dict)

        # Query again using 'next' hint
        url_param = next.split('/')[-1]
        resp, body = self.client.get_list_obj(url_param)

        self.assertEqual(200, resp.status)
        self.assertEqual(1, len(body['actions']))

        name_2 = body['actions'][0].get('name')

        self.assertGreater(name_1, name_2)

    @test.attr(type='negative')
    def test_get_list_actions_nonexist_sort_dirs(self):
        context = self.assertRaises(
            exceptions.BadRequest,
            self.client.get_list_obj,
            'actions?limit=1&sort_keys=id&sort_dirs=nonexist'
        )

        self.assertIn(
            'Unknown sort direction',
            context.resp_body.get('faultstring')
        )

    @test.attr(type='negative')
    def test_get_list_actions_invalid_limit(self):
        context = self.assertRaises(
            exceptions.BadRequest,
            self.client.get_list_obj,
            'actions?limit=-1&sort_keys=id&sort_dirs=asc'
        )

        self.assertIn(
            'Limit must be positive',
            context.resp_body.get('faultstring')
        )

    @test.attr(type='negative')
    def test_get_list_actions_duplicate_sort_keys(self):
        context = self.assertRaises(
            exceptions.BadRequest,
            self.client.get_list_obj,
            'actions?limit=1&sort_keys=id,id&sort_dirs=asc,asc'
        )

        self.assertIn(
            'Length of sort_keys must be equal or greater than sort_dirs',
            context.resp_body.get('faultstring')
        )

    @test.attr(type='smoke')
    def test_get_list_actions_equal_to_filter(self):
        resp, body = self.client.create_action('action_v2.yaml')
        self.assertEqual(201, resp.status)

        resp, body = self.client.get_list_obj(
            'actions?is_system=False'
        )

        self.assertEqual(200, resp.status)
        self.assertNotEqual([], body['actions'])

        for act in body['actions']:
            self.assertFalse(act['is_system'])

    @test.attr(type='smoke')
    def test_get_list_actions_not_equal_to_filter(self):
        resp, body = self.client.create_action('action_v2.yaml')
        self.assertEqual(201, resp.status)

        resp, body = self.client.get_list_obj(
            'actions?is_system=neq:False'
        )

        self.assertEqual(200, resp.status)
        self.assertNotEqual([], body['actions'])

        for act in body['actions']:
            self.assertTrue(act['is_system'])

    @test.attr(type='smoke')
    def test_get_list_actions_in_list_filter(self):
        resp, body = self.client.create_action('action_v2.yaml')
        self.assertEqual(201, resp.status)

        created_acts = [action['name'] for action in body['actions']]
        _, body = self.client.get_object('actions', created_acts[0])
        time = body['created_at']
        resp, body = self.client.get_list_obj(
            'actions?created_at=in:' + time.replace(' ', '%20')
        )

        self.assertEqual(200, resp.status)
        action_names = [action['name'] for action in body['actions']]
        self.assertListEqual(created_acts, action_names)

    @test.attr(type='smoke')
    def test_get_list_actions_not_in_list_filter(self):
        resp, body = self.client.create_action('action_v2.yaml')
        self.assertEqual(201, resp.status)

        created_acts = [action['name'] for action in body['actions']]
        _, body = self.client.get_object('actions', created_acts[0])
        time = body['created_at']
        resp, body = self.client.get_list_obj(
            'actions?created_at=nin:' + time.replace(' ', '%20')
        )

        self.assertEqual(200, resp.status)
        action_names = [action['name'] for action in body['actions']]
        for act in created_acts:
            self.assertNotIn(act, action_names)

    @test.attr(type='smoke')
    def test_get_list_actions_greater_than_filter(self):
        time = datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")
        resp, body = self.client.get_list_obj(
            'actions?created_at=gt:' + time.replace(' ', '%20')
        )

        self.assertEqual(200, resp.status)
        self.assertEqual([], body['actions'])

    @test.attr(type='smoke')
    def test_get_list_actions_greater_than_equal_to_filter(self):
        resp, body = self.client.create_action('action_v2.yaml')
        self.assertEqual(201, resp.status)

        created_acts = [action['name'] for action in body['actions']]
        _, body = self.client.get_object('actions', created_acts[0])
        time = body['created_at']
        resp, body = self.client.get_list_obj(
            'actions?created_at=gte:' + time.replace(' ', '%20')
        )

        actions = [action['name'] for action in body['actions']]
        self.assertEqual(200, resp.status)
        self.assertIn(created_acts[0], actions)

    @test.attr(type='smoke')
    def test_get_list_actions_less_than_filter(self):
        resp, body = self.client.create_action('action_v2.yaml')
        self.assertEqual(201, resp.status)

        created_acts = [action['name'] for action in body['actions']]
        _, body = self.client.get_object('actions', created_acts[0])
        time = body['created_at']
        resp, body = self.client.get_list_obj(
            'actions?created_at=lt:' + time.replace(' ', '%20')
        )

        actions = [action['name'] for action in body['actions']]
        self.assertEqual(200, resp.status)
        self.assertNotIn(created_acts[0], actions)

    @test.attr(type='smoke')
    def test_get_list_actions_less_than_equal_to_filter(self):
        resp, body = self.client.create_action('action_v2.yaml')
        self.assertEqual(201, resp.status)

        created_acts = [action['name'] for action in body['actions']]
        _, body = self.client.get_object('actions', created_acts[0])
        time = body['created_at']
        resp, body = self.client.get_list_obj(
            'actions?created_at=lte:' + time.replace(' ', '%20')
        )

        actions = [action['name'] for action in body['actions']]
        self.assertEqual(200, resp.status)
        self.assertIn(created_acts[0], actions)

    @test.attr(type='smoke')
    def test_get_list_actions_multiple_filter(self):
        resp, body = self.client.create_action('action_v2.yaml')
        self.assertEqual(201, resp.status)

        created_acts = [action['name'] for action in body['actions']]
        _, body = self.client.get_object('actions', created_acts[0])
        time = body['created_at']
        resp, body = self.client.get_list_obj(
            'actions?created_at=lte:' + time.replace(' ', '%20') +
            '&is_system=False'
        )

        actions = [action['name'] for action in body['actions']]
        self.assertEqual(200, resp.status)
        self.assertIn(created_acts[0], actions)

    @test.attr(type='negative')
    def test_get_list_actions_invalid_filter(self):
        self.assertRaises(
            exceptions.BadRequest,
            self.client.get_list_obj,
            'actions?is_system<False'
        )
        self.assertRaises(
            exceptions.BadRequest,
            self.client.get_list_obj,
            'actions?is_system!=False'
        )
        self.assertRaises(
            exceptions.BadRequest,
            self.client.get_list_obj,
            'actions?created_at>2016-02-23%2008:51:26'
        )

    @test.attr(type='sanity')
    def test_create_and_delete_few_actions(self):
        resp, body = self.client.create_action('action_v2.yaml')
        self.assertEqual(201, resp.status)

        created_acts = [action['name'] for action in body['actions']]

        resp, body = self.client.get_list_obj('actions')
        self.assertEqual(200, resp.status)

        actions = [action['name'] for action in body['actions']]

        for act in created_acts:
            self.assertIn(act, actions)
            self.client.delete_obj('actions', act)

        _, body = self.client.get_list_obj('actions')
        actions = [action['name'] for action in body['actions']]

        for act in created_acts:
            self.assertNotIn(act, actions)
            self.client.actions.remove(act)

    @test.attr(type='sanity')
    def test_get_action(self):
        _, body = self.client.create_action('action_v2.yaml')
        action_name = body['actions'][0]['name']
        resp, body = self.client.get_object('actions', action_name)

        self.assertEqual(200, resp.status)
        self.assertEqual(action_name, body['name'])

    @test.attr(type='sanity')
    def test_update_action(self):
        _, body = self.client.create_action('action_v2.yaml')
        action = body['actions'][0]['name']

        act_created_at = self.get_field_value(
            body=body, act_name=action, field='created_at')

        self.assertNotIn('updated at', body['actions'])

        resp, body = self.client.update_request('actions', 'action_v2.yaml')
        self.assertEqual(200, resp.status)

        actions = [act['name'] for act in body['actions']]
        self.assertIn(action, actions)

        updated_act_created_at = self.get_field_value(
            body=body, act_name=action, field='created_at')

        self.assertEqual(act_created_at.split(".")[0], updated_act_created_at)
        self.assertTrue(all(['updated_at' in item
                             for item in body['actions']]))

    @test.attr(type='sanity')
    def test_get_action_definition(self):
        _, body = self.client.create_action('action_v2.yaml')
        act_name = body['actions'][0]['name']

        resp, body = self.client.get_definition('actions', act_name)
        self.assertEqual(200, resp.status)
        self.assertIsNotNone(body)
        self.assertIn(act_name, body)

    @test.attr(type='negative')
    def test_get_nonexistent_action(self):
        self.assertRaises(
            exceptions.NotFound,
            self.client.get_object,
            'actions', 'nonexist'
        )

    @test.attr(type='negative')
    def test_double_creation(self):
        self.client.create_action('action_v2.yaml')

        self.assertRaises(
            exceptions.Conflict,
            self.client.create_action,
            'action_v2.yaml'
        )

    @test.attr(type='negative')
    def test_create_action_invalid_def(self):
        self.assertRaises(
            exceptions.BadRequest,
            self.client.create_action,
            'wb_v2.yaml'
        )

    @test.attr(type='negative')
    def test_update_action_invalid_def(self):
        self.assertRaises(
            exceptions.BadRequest,
            self.client.update_request,
            'actions', 'wb_v2.yaml'
        )

    @test.attr(type='negative')
    def test_delete_nonexistent_action(self):
        self.assertRaises(
            exceptions.NotFound,
            self.client.delete_obj,
            'actions', 'nonexist'
        )

    @test.attr(type='negative')
    def test_delete_standard_action(self):
        self.assertRaises(
            exceptions.BadRequest,
            self.client.delete_obj,
            'actions', 'nova.servers_create'
        )



# Copyright 2013 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import copy
import datetime
import json
import mock

from mistral.db.v2 import api as db_api
from mistral.db.v2.sqlalchemy import models
from mistral.engine.rpc_backend import rpc
from mistral import exceptions as exc
from mistral.tests.unit.api import base
from mistral.workflow import data_flow
from mistral.workflow import states

# TODO(everyone): later we need additional tests verifying all the errors etc.

RESULT = {"some": "result"}
PUBLISHED = {"var": "val"}

WF_EX = models.WorkflowExecution(
    id='abc',
    workflow_name='some',
    description='execution description.',
    spec={'name': 'some'},
    state=states.RUNNING,
    state_info=None,
    input={'foo': 'bar'},
    output={},
    params={'env': {'k1': 'abc'}},
    created_at=datetime.datetime(1970, 1, 1),
    updated_at=datetime.datetime(1970, 1, 1)
)

TASK_EX = models.TaskExecution(
    id='123',
    name='task',
    workflow_name='flow',
    workflow_id='123e4567-e89b-12d3-a456-426655441111',
    spec={
        'type': 'direct',
        'version': '2.0',
        'name': 'task'
    },
    action_spec={},
    state=states.RUNNING,
    tags=['a', 'b'],
    in_context={},
    runtime_context={},
    workflow_execution_id=WF_EX.id,
    created_at=datetime.datetime(1970, 1, 1),
    updated_at=datetime.datetime(1970, 1, 1),
    published=PUBLISHED,
    processed=True
)

WITH_ITEMS_TASK_EX = models.TaskExecution(
    id='123',
    name='task',
    workflow_name='flow',
    workflow_id='123e4567-e89b-12d3-a456-426655441111',
    spec={
        'type': 'direct',
        'version': '2.0',
        'name': 'task',
        'with-items': 'var in [1, 2, 3]'
    },
    action_spec={},
    state=states.RUNNING,
    tags=['a', 'b'],
    in_context={},
    runtime_context={},
    workflow_execution_id=WF_EX.id,
    created_at=datetime.datetime(1970, 1, 1),
    updated_at=datetime.datetime(1970, 1, 1),
    published=PUBLISHED,
    processed=True
)

TASK = {
    'id': '123',
    'name': 'task',
    'workflow_name': 'flow',
    'workflow_id': '123e4567-e89b-12d3-a456-426655441111',
    'state': 'RUNNING',
    'workflow_execution_id': WF_EX.id,
    'created_at': '1970-01-01 00:00:00',
    'updated_at': '1970-01-01 00:00:00',
    'result': json.dumps(RESULT),
    'published': json.dumps(PUBLISHED),
    'processed': True
}

TASK_WITHOUT_RESULT = copy.deepcopy(TASK)
del TASK_WITHOUT_RESULT['result']

UPDATED_TASK_EX = copy.deepcopy(TASK_EX)
UPDATED_TASK_EX['state'] = 'SUCCESS'
UPDATED_TASK = copy.deepcopy(TASK)
UPDATED_TASK['state'] = 'SUCCESS'

ERROR_TASK_EX = copy.deepcopy(TASK_EX)
ERROR_TASK_EX['state'] = 'ERROR'
ERROR_ITEMS_TASK_EX = copy.deepcopy(WITH_ITEMS_TASK_EX)
ERROR_ITEMS_TASK_EX['state'] = 'ERROR'
ERROR_TASK = copy.deepcopy(TASK)
ERROR_TASK['state'] = 'ERROR'

BROKEN_TASK = copy.deepcopy(TASK)

RERUN_TASK = {
    'id': '123',
    'state': 'RUNNING'
}

MOCK_WF_EX = mock.MagicMock(return_value=WF_EX)
MOCK_TASK = mock.MagicMock(return_value=TASK_EX)
MOCK_TASKS = mock.MagicMock(return_value=[TASK_EX])
MOCK_EMPTY = mock.MagicMock(return_value=[])
MOCK_NOT_FOUND = mock.MagicMock(side_effect=exc.DBEntityNotFoundError())
MOCK_ERROR_TASK = mock.MagicMock(return_value=ERROR_TASK_EX)
MOCK_ERROR_ITEMS_TASK = mock.MagicMock(return_value=ERROR_ITEMS_TASK_EX)


@mock.patch.object(
    data_flow,
    'get_task_execution_result', mock.Mock(return_value=RESULT)
)
class TestTasksController(base.APITest):
    @mock.patch.object(db_api, 'get_task_execution', MOCK_TASK)
    def test_get(self):
        resp = self.app.get('/v2/tasks/123')

        self.assertEqual(200, resp.status_int)
        self.assertDictEqual(TASK, resp.json)

    @mock.patch.object(db_api, 'get_task_execution', MOCK_NOT_FOUND)
    def test_get_not_found(self):
        resp = self.app.get('/v2/tasks/123', expect_errors=True)

        self.assertEqual(404, resp.status_int)

    @mock.patch.object(db_api, 'get_task_executions', MOCK_TASKS)
    def test_get_all(self):
        resp = self.app.get('/v2/tasks')

        self.assertEqual(200, resp.status_int)

        self.assertEqual(1, len(resp.json['tasks']))
        self.assertDictEqual(TASK_WITHOUT_RESULT, resp.json['tasks'][0])

    @mock.patch.object(db_api, 'get_task_executions', MOCK_EMPTY)
    def test_get_all_empty(self):
        resp = self.app.get('/v2/tasks')

        self.assertEqual(200, resp.status_int)

        self.assertEqual(0, len(resp.json['tasks']))

    @mock.patch.object(db_api, 'get_workflow_execution', MOCK_WF_EX)
    @mock.patch.object(
        db_api,
        'get_task_execution',
        mock.MagicMock(side_effect=[ERROR_TASK_EX, TASK_EX])
    )
    @mock.patch.object(rpc.EngineClient, 'rerun_workflow', MOCK_WF_EX)
    def test_put(self):
        params = copy.deepcopy(RERUN_TASK)
        params['reset'] = True

        resp = self.app.put_json('/v2/tasks/123', params=params)

        self.assertEqual(200, resp.status_int)
        self.assertDictEqual(TASK, resp.json)

        rpc.EngineClient.rerun_workflow.assert_called_with(
            TASK_EX.id,
            reset=params['reset'],
            env=None
        )

    @mock.patch.object(db_api, 'get_workflow_execution', MOCK_WF_EX)
    @mock.patch.object(
        db_api,
        'get_task_execution',
        mock.MagicMock(side_effect=[ERROR_TASK_EX, TASK_EX])
    )
    @mock.patch.object(rpc.EngineClient, 'rerun_workflow', MOCK_WF_EX)
    def test_put_missing_reset(self):
        params = copy.deepcopy(RERUN_TASK)

        resp = self.app.put_json(
            '/v2/tasks/123',
            params=params,
            expect_errors=True)

        self.assertEqual(400, resp.status_int)
        self.assertIn('faultstring', resp.json)
        self.assertIn('Mandatory field missing', resp.json['faultstring'])

    @mock.patch.object(db_api, 'get_workflow_execution', MOCK_WF_EX)
    @mock.patch.object(
        db_api,
        'get_task_execution',
        mock.MagicMock(side_effect=[ERROR_ITEMS_TASK_EX, WITH_ITEMS_TASK_EX])
    )
    @mock.patch.object(rpc.EngineClient, 'rerun_workflow', MOCK_WF_EX)
    def test_put_with_items(self):
        params = copy.deepcopy(RERUN_TASK)
        params['reset'] = False

        resp = self.app.put_json('/v2/tasks/123', params=params)

        self.assertEqual(200, resp.status_int)
        self.assertDictEqual(TASK, resp.json)

    @mock.patch.object(db_api, 'get_workflow_execution', MOCK_WF_EX)
    @mock.patch.object(
        db_api,
        'get_task_execution',
        mock.MagicMock(side_effect=[ERROR_TASK_EX, TASK_EX])
    )
    @mock.patch.object(rpc.EngineClient, 'rerun_workflow', MOCK_WF_EX)
    def test_put_env(self):
        params = copy.deepcopy(RERUN_TASK)
        params['reset'] = True
        params['env'] = '{"k1": "def"}'

        resp = self.app.put_json('/v2/tasks/123', params=params)

        self.assertEqual(200, resp.status_int)
        self.assertDictEqual(TASK, resp.json)

        rpc.EngineClient.rerun_workflow.assert_called_with(
            TASK_EX.id,
            reset=params['reset'],
            env=json.loads(params['env'])
        )

    @mock.patch.object(db_api, 'get_workflow_execution', MOCK_WF_EX)
    @mock.patch.object(db_api, 'get_task_execution', MOCK_TASK)
    def test_put_current_task_not_in_error(self):
        params = copy.deepcopy(RERUN_TASK)
        params['reset'] = True

        resp = self.app.put_json(
            '/v2/tasks/123',
            params=params,
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)
        self.assertIn('faultstring', resp.json)
        self.assertIn('execution must be in ERROR', resp.json['faultstring'])

    @mock.patch.object(rpc.EngineClient, 'rerun_workflow', MOCK_WF_EX)
    @mock.patch.object(db_api, 'get_workflow_execution', MOCK_WF_EX)
    @mock.patch.object(db_api, 'get_task_execution', MOCK_ERROR_TASK)
    def test_put_current_task_in_error(self):
        params = copy.deepcopy(RERUN_TASK)
        params['reset'] = True
        params['env'] = '{"k1": "def"}'

        resp = self.app.put_json('/v2/tasks/123', params=params)

        self.assertEqual(200, resp.status_int)

    @mock.patch.object(db_api, 'get_workflow_execution', MOCK_WF_EX)
    @mock.patch.object(db_api, 'get_task_execution', MOCK_ERROR_TASK)
    def test_put_invalid_state(self):
        params = copy.deepcopy(RERUN_TASK)
        params['state'] = states.IDLE
        params['reset'] = True

        resp = self.app.put_json(
            '/v2/tasks/123',
            params=params,
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)
        self.assertIn('faultstring', resp.json)
        self.assertIn('Invalid task state', resp.json['faultstring'])

    @mock.patch.object(db_api, 'get_workflow_execution', MOCK_WF_EX)
    @mock.patch.object(db_api, 'get_task_execution', MOCK_ERROR_TASK)
    def test_put_invalid_reset(self):
        params = copy.deepcopy(RERUN_TASK)
        params['reset'] = False

        resp = self.app.put_json(
            '/v2/tasks/123',
            params=params,
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)
        self.assertIn('faultstring', resp.json)
        self.assertIn('Only with-items task', resp.json['faultstring'])

        @mock.patch.object(db_api, 'get_workflow_execution', MOCK_WF_EX)
        @mock.patch.object(db_api, 'get_task_execution', MOCK_ERROR_TASK)
        def test_put_valid_state(self):
            params = copy.deepcopy(RERUN_TASK)
            params['state'] = states.RUNNING
            params['reset'] = True

            resp = self.app.put_json(
                '/v2/tasks/123',
                params=params
            )

            self.assertEqual(200, resp.status_int)

    @mock.patch.object(db_api, 'get_workflow_execution', MOCK_WF_EX)
    @mock.patch.object(db_api, 'get_task_execution', MOCK_ERROR_TASK)
    def test_put_mismatch_task_name(self):
        params = copy.deepcopy(RERUN_TASK)
        params['name'] = 'abc'
        params['reset'] = True

        resp = self.app.put_json(
            '/v2/tasks/123',
            params=params,
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)
        self.assertIn('faultstring', resp.json)
        self.assertIn('Task name does not match', resp.json['faultstring'])

    @mock.patch.object(rpc.EngineClient, 'rerun_workflow', MOCK_WF_EX)
    @mock.patch.object(db_api, 'get_workflow_execution', MOCK_WF_EX)
    @mock.patch.object(db_api, 'get_task_execution', MOCK_ERROR_TASK)
    def test_put_match_task_name(self):
        params = copy.deepcopy(RERUN_TASK)
        params['name'] = 'task'
        params['reset'] = True

        resp = self.app.put_json(
            '/v2/tasks/123',
            params=params,
            expect_errors=True
        )

        self.assertEqual(200, resp.status_int)

    @mock.patch.object(db_api, 'get_workflow_execution', MOCK_WF_EX)
    @mock.patch.object(db_api, 'get_task_execution', MOCK_ERROR_TASK)
    def test_put_mismatch_workflow_name(self):
        params = copy.deepcopy(RERUN_TASK)
        params['workflow_name'] = 'xyz'
        params['reset'] = True

        resp = self.app.put_json(
            '/v2/tasks/123',
            params=params,
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)
        self.assertIn('faultstring', resp.json)
        self.assertIn('Workflow name does not match', resp.json['faultstring'])
